# -*- coding: UTF-8 -*-

import numpy as np

from lava.api.constants.spirv import DataType, Layout, Order
from lava.api.util import NdArray, LavaError


class BytesError(LavaError):

    def __init__(self, message):
        super(BytesError, self).__init__(message)

    @classmethod
    def out_of_bounds(cls, value, glsl_dtype):
        raise cls("Value {} is out of memory bounds for type {}".format(value, glsl_dtype))


class ByteRepresentation(object):

    """Data structure implementation of Vulkan specification 14.5.4."""

    # References
    # https://www.khronos.org/registry/vulkan/specs/1.1/html/chap14.html#interfaces-resources-layout
    # https://www.khronos.org/registry/OpenGL/extensions/ARB/ARB_uniform_buffer_object.txt
    #   search document for "Sub-section 2.15.3.1.2"
    #   search document for "The following example illustrates the rules specified by the "std140" layout"

    def __init__(self):
        pass

    def copy(self):
        raise NotImplementedError()

    def size(self):
        raise NotImplementedError()

    def alignment(self):
        # called 'base alignment' in the specs
        raise NotImplementedError()

    def __str__(self, indent=2):
        raise NotImplementedError()

    # def glsl(self, var_name):
    #     return "{} {};".format(self.glsl_dtype(), var_name)

    def glsl_dtype(self):
        raise NotImplementedError()

    @classmethod
    def path_to_str(cls, path):
        return " > ".join(path)

    @classmethod
    def compare_type(cls, type_expected, type_other, path, quiet=True):
        if type_expected != type_other:
            if not quiet:
                raise TypeError(
                    "Expected type {}, but got type {} at {}".format(type_expected, type_other, cls.path_to_str(path)))
            return False

        return True

    @classmethod
    def compare_layout(cls, layout_expected, layout_other, path, quiet=True):
        if layout_expected == Layout.STDXXX:
            return True

        if layout_expected != layout_other and layout_other != Layout.STDXXX:
            if not quiet:
                raise TypeError(
                    "Expected layout {}, but got layout {} at {}".format(layout_expected, layout_other,
                                                                         cls.path_to_str(path)))
            return False

        return True

    @classmethod
    def compare_order(cls, order_expected, order_other, path, quiet=True):
        if order_expected != order_other:
            if not quiet:
                raise TypeError(
                    "Expected layout {}, but got layout {} at {}".format(order_expected, order_other,
                                                                         cls.path_to_str(path)))
            return False
        return True

    @classmethod
    def compare_shape(cls, shape_expected, shape_other, path, quiet):
        if shape_expected != shape_other:
            if not quiet:
                raise TypeError("Expected shape {}, but got shape {} at {}".format(shape_expected, shape_other,
                                                                                   cls.path_to_str(path)))
            return False
        return True

    def compare(self, other, path, quiet=True):
        raise NotImplementedError()

    def to_bytes(self, values, path=()):
        raise NotImplementedError()

    def from_bytes(self, bytez):
        raise NotImplementedError()


class Scalar(ByteRepresentation):

    def __init__(self, input_dtypes):
        super(Scalar, self).__init__()
        self.input_dtypes = input_dtypes

    @classmethod
    def int(cls):
        return cls.of(DataType.INT)

    @classmethod
    def uint(cls):
        return cls.of(DataType.UINT)

    @classmethod
    def float(cls):
        return cls.of(DataType.FLOAT)

    @classmethod
    def double(cls):
        return cls.of(DataType.DOUBLE)

    @classmethod
    def of(cls, dtype):
        if dtype == DataType.INT:
            return ScalarInt()
        if dtype == DataType.UINT:
            return ScalarUnsignedInt()
        if dtype == DataType.FLOAT:
            return ScalarFloat()
        if dtype == DataType.DOUBLE:
            return ScalarDouble()
        raise ValueError("Unknown scalar type '{}'".format(dtype))

    def copy(self):
        return self.__class__()

    def alignment(self):
        # "A scalar of size N has a base alignment of N."
        return self.size()

    def __str__(self, name=None, indent=2):
        return "{} [{}]".format(self.glsl_dtype(), name or "?")

    def glsl_dtype(self):
        raise NotImplementedError()

    def compare(self, other, path=(), quiet=True):
        return self.compare_type(type(self), type(other), path, quiet)

    def to_bytes(self, value, path=()):
        if not isinstance(value, self.input_dtypes):
            raise TypeError("{} got dtype {}, expected one of the following {} at {}".format(
                self.__class__.__name__, type(value), self.input_dtypes, self.path_to_str(path)
            ))

    def from_bytes(self, bytez):
        return np.frombuffer(bytez, self.numpy_dtype())[0]

    def numpy_dtype(self):
        raise NotImplementedError()


class ScalarInt(Scalar):

    def __init__(self):
        super(ScalarInt, self).__init__((int, np.int32))

    def numpy_dtype(self):
        return np.int32

    def size(self):
        return 4

    def glsl_dtype(self):
        return "int"

    def to_bytes(self, value, path=()):
        super(ScalarInt, self).to_bytes(value, path)
        if type(value) is int:
            if not (-(0x7FFFFFFF + 1) <= value <= 0x7FFFFFFF):
                raise BytesError.out_of_bounds(value, self.glsl_dtype())
            value = np.int32(value)
        return bytearray(value.tobytes())


class ScalarUnsignedInt(Scalar):

    def __init__(self):
        super(ScalarUnsignedInt, self).__init__((int, np.uint32, bool, np.bool, np.bool_))

    def numpy_dtype(self):
        return np.uint32

    def size(self):
        return 4

    def glsl_dtype(self):
        return "uint"

    def to_bytes(self, value, path=()):
        super(ScalarUnsignedInt, self).to_bytes(value, path)
        if not (0 <= value <= 0xFFFFFFFF):
            raise BytesError.out_of_bounds(value, self.glsl_dtype())
        if type(value) != np.uint32:
            value = np.uint32(value)
        return bytearray(value.tobytes())


class ScalarFloat(Scalar):

    def __init__(self):
        super(ScalarFloat, self).__init__((float, np.float32))

    def numpy_dtype(self):
        return np.float32

    def size(self):
        return 4

    def glsl_dtype(self):
        return "float"

    def to_bytes(self, value, path=()):
        super(ScalarFloat, self).to_bytes(value, path)
        if type(value) is float:
            # TODO: add range check
            value = np.float32(value)
        return bytearray(value.tobytes())


class ScalarDouble(Scalar):

    def __init__(self):
        super(ScalarDouble, self).__init__((float, np.float64))

    def numpy_dtype(self):
        return np.float64

    def size(self):
        return 8

    def glsl_dtype(self):
        return "double"

    def to_bytes(self, value, path=()):
        super(ScalarDouble, self).to_bytes(value, path)
        if type(value) is float:
            # TODO: add range check
            value = np.float64(value)
        return bytearray(value.tobytes())


class Vector(ByteRepresentation):

    def __init__(self, n, dtype):
        super(Vector, self).__init__()
        self.dtype = dtype
        self.n = n
        self.scalar = Scalar.of(dtype)

    @classmethod
    def ivec2(cls):
        return Vector(2, DataType.INT)

    @classmethod
    def ivec3(cls):
        return Vector(3, DataType.INT)

    @classmethod
    def ivec4(cls):
        return Vector(4, DataType.INT)

    @classmethod
    def uvec2(cls):
        return Vector(2, DataType.UINT)

    @classmethod
    def uvec3(cls):
        return Vector(3, DataType.UINT)

    @classmethod
    def uvec4(cls):
        return Vector(4, DataType.UINT)

    @classmethod
    def vec2(cls):
        return Vector(2, DataType.FLOAT)

    @classmethod
    def vec3(cls):
        return Vector(3, DataType.FLOAT)

    @classmethod
    def vec4(cls):
        return Vector(4, DataType.FLOAT)

    @classmethod
    def dvec2(cls):
        return Vector(2, DataType.DOUBLE)

    @classmethod
    def dvec3(cls):
        return Vector(3, DataType.DOUBLE)

    @classmethod
    def dvec4(cls):
        return Vector(4, DataType.DOUBLE)

    def copy(self):
        return Vector(self.n, self.dtype)

    def size(self):
        return self.scalar.size() * self.n

    def length(self):
        return self.n

    def alignment(self):
        if self.n == 2:
            # "A two-component vector, with components of size N, has a base alignment of 2 N."
            return self.scalar.size() * 2
        if self.n in (3, 4):
            # "A three- or four-component vector, with components of size N, has a base alignment of 4 N."
            return self.scalar.size() * 4
        return -1

    def __str__(self, name=None, indent=2):
        return "{} [{}]".format(self.glsl_dtype(), name or "?")

    def glsl_dtype(self):
        return "{}vec{}".format(self.dtype.lower()[0] if self.dtype is not DataType.FLOAT else "", self.n)

    def compare(self, other, path=(), quiet=True):
        if not self.compare_type(type(self), type(other), path, quiet):
            return False

        count_expected = self.n
        count_other = other.n

        if count_expected != count_other:
            if not quiet:
                raise TypeError("Expected vector length {}, but got length {} at {}".format(
                    count_expected, count_other, self.path_to_str(path)))
            return False

        return True

    def to_bytes(self, array, path=()):
        expected_types = (np.ndarray, list, tuple)
        if not isinstance(array, expected_types):
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(type(array), self.glsl_dtype(), expected_types, self.path_to_str(path)))
        if len(array) != self.n:
            raise TypeError("Got length {} for {} variable, expected {} at {}"
                            .format(len(array), self.glsl_dtype(), self.n, self.path_to_str(path)))

        bytez = bytearray()

        for value in array:
            bytez += self.scalar.to_bytes(value, path)

        return bytez

    def from_bytes(self, bytez):
        return np.frombuffer(bytez, self.scalar.numpy_dtype())[:self.n]


class Matrix(ByteRepresentation):

    DEFAULT_ORDER = Order.COLUMN_MAJOR

    def __init__(self, cols, rows, dtype, layout, order=DEFAULT_ORDER):
        super(Matrix, self).__init__()
        if dtype not in (DataType.FLOAT, DataType.DOUBLE):
            raise TypeError("Matrices of type {} are not supported".format(dtype))
        self.dtype = dtype
        self.cols = cols
        self.rows = rows
        self.order = order
        self.layout = layout

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, order):
        self._order = order
        self.vector = Vector(self.rows if order == Order.COLUMN_MAJOR else self.cols, self.dtype)
        self.vector_count = self.cols if order == Order.COLUMN_MAJOR else self.rows

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        self._layout = layout
        self.a = None
        self.precompute_alignment()

    def precompute_alignment(self):
        if self.layout in [Layout.STD140, Layout.STD430, Layout.STDXXX]:
            self.a = self.vector.alignment()

            if self.layout == Layout.STD140:
                self.a += (16 - self.a % 16) % 16

    def copy(self):
        return Matrix(self.cols, self.rows, self.dtype, self.layout, self.order)

    def size(self):
        return self.step_size() * self.vector_count

    def stride(self):
        return self.step_size()

    def step_size(self):
        s = self.vector.size()
        s += (self.a - s % self.a) % self.a  # pad to array stride
        return s

    def shape(self):
        return self.rows, self.cols

    def alignment(self):
        return self.a

    def __str__(self, name=None, indent=2):
        return "{} [{}]".format(self.glsl_dtype(), name or "?")

    def glsl_dtype(self):
        return "{}mat{}x{}".format(self.dtype.lower()[0] if self.dtype is not DataType.FLOAT else "", self.cols, self.rows)

    def compare(self, other, path=(), quiet=True):
        if not self.compare_type(type(self), type(other), path, quiet):
            return False
        if not self.compare_layout(self.layout, other.layout, path, quiet):
            return False
        if not self.compare_shape(self.shape(), other.shape(), path, quiet):
            return False
        return True

    def to_bytes(self, array, path=()):
        expected_types = (np.ndarray, list, tuple)
        if not (isinstance(array, expected_types) or isinstance(array[0], expected_types)):
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(type(array), self.glsl_dtype(), expected_types, self.path_to_str(path)))
        shape = (len(array), len(array[0]))
        expected_shape = self.shape()
        if shape != expected_shape:
            raise TypeError("Got shape {} for {} variable, expected {} at {}"
                            .format(len(array), self.glsl_dtype(), expected_shape, self.path_to_str(path)))

        bytez = bytearray()

        if self.order == Order.COLUMN_MAJOR:
            for col in range(self.cols):
                bytez += self.vector.to_bytes([array[row][col] for row in range(self.rows)], path)
                padding = (self.a - len(bytez) % self.a) % self.a
                bytez += bytearray(padding)

        if self.order == Order.ROW_MAJOR:
            for row in range(self.rows):
                bytez += self.vector.to_bytes([array[row][col] for col in range(self.cols)], path)
                padding = (self.a - len(bytez) % self.a) % self.a
                bytez += bytearray(padding)

        return bytez

    def from_bytes(self, bytez):
        array = np.zeros((self.rows, self.cols), dtype=self.vector.scalar.numpy_dtype())
        offset = 0
        size = self.vector.size()

        if self.order == Order.COLUMN_MAJOR:
            for col in range(self.cols):
                array[:, col] = self.vector.from_bytes(bytez[offset:offset + size])
                offset += size
                offset += (self.a - offset % self.a) % self.a  # bytes

            return array

        if self.order == Order.ROW_MAJOR:
            for row in range(self.rows):
                array[row, :] = self.vector.from_bytes(bytez[offset:offset + size])
                offset += size
                offset += (self.a - offset % self.a) % self.a  # bytes

            return array

        return None


class Array(ByteRepresentation):

    def __init__(self, definition, dims, layout):
        super(Array, self).__init__()
        self.definition = definition
        self.dims = tuple(dims) if isinstance(dims, (list, tuple)) else (dims,)
        self.layout = layout  # precomputes alignment

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        self._layout = layout

        # set children first
        if isinstance(self.definition, (Matrix, Struct)):
            self.definition.layout = layout

        self.a = None
        self.precompute_alignment()

    def precompute_alignment(self):
        if self.layout in [Layout.STD140, Layout.STD430, Layout.STDXXX]:
            self.a = self.definition.alignment()

            if self.layout == Layout.STD140:
                self.a += (16 - self.a % 16) % 16

    def shape(self):
        return self.dims

    def copy(self):
        return Array(self.definition.copy(), self.dims, self.layout)

    def size(self):
        return self.step_size() * np.product(self.shape())

    def strides(self):
        strides = [self.step_size()]
        shape = self.shape()

        for i in reversed(range(1, len(shape))):
            strides.insert(0, shape[i] * strides[0])

        return strides

    def step_size(self):
        s = self.definition.size()
        s += (self.a - s % self.a) % self.a  # pad to array stride
        return s

    def alignment(self):
        return self.a

    @classmethod
    def is_array_of_structs(cls, definition):
        if isinstance(definition, Array):
            if isinstance(definition.definition, Struct):
                return True
        return False

    def __str__(self, name=None, indent=2):
        s = self.definition.glsl_dtype()  # "array"
        s += ("[{}]" * len(self.shape())).format(*self.shape())
        s += " [{}] {{ {} }}".format(name or "?", self.definition.__str__(indent=indent + 2))
        return s

    def glsl_dtype(self):
        return ("{}" + "[{}]" * len(self.shape())).format(self.definition.glsl_dtype(), *self.shape())

    def compare(self, other, path=(), quiet=True):
        if not self.compare_type(type(self), type(other), path, quiet):
            return False
        if not self.compare_layout(self.layout, other.layout, path, quiet):
            return False
        if not self.compare_shape(self.shape(), other.shape(), path, quiet):
            return False

        return self.definition.compare(other.definition,
                                       list(path) + ["array {}".format("x".join(map(str, self.shape())))], quiet)

    def to_bytes(self, values, path=()):
        if isinstance(self.definition, Scalar):
            return self.to_bytes_for_scalars(values, path)

        elif isinstance(self.definition, Vector):
            return self.to_bytes_for_vectors(values, path)

        elif isinstance(self.definition, Matrix):
            return self.to_bytes_for_matrices(values, path)

        else:
            bytez = bytearray()

            for indices in NdArray.iterate(self.shape()):
                bytez += self.definition.to_bytes(NdArray.get(values, indices),
                                                  list(path) + ["array" + "".join(["[{}]".format(i) for i in indices])])
                padding = (self.a - len(bytez) % self.a) % self.a
                bytez += bytearray(padding)

            return bytez

    def to_bytes_for_scalars(self, array, path):
        transfer_dtype = self.definition.numpy_dtype()

        if not isinstance(array, np.ndarray):
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(type(array), self.glsl_dtype(), np.ndarray, self.path_to_str(path)))
        if array.dtype.type not in self.definition.input_dtypes:
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(array.dtype, self.glsl_dtype(), self.definition.input_dtypes, self.path_to_str(path)))
        if tuple(array.shape) != self.shape():
            raise TypeError("Got shape {} for {} variable, expected {} at {}"
                            .format(array.shape, self.glsl_dtype(), self.shape(), self.path_to_str(path)))

        if self.layout == Layout.STD430:
            return array.astype(transfer_dtype).flatten().tobytes()

        else:
            p = (self.a - self.definition.alignment()) // self.definition.size()
            a = self.a // self.definition.size()

            array_padded = np.zeros(a * np.product(array.shape), dtype=transfer_dtype)
            mask = (np.arange(len(array_padded)) % a) < (a - p)
            array_padded[mask] = array.flatten()

            return array_padded.tobytes()

    def to_bytes_for_vectors(self, array, path):
        numpy_dtypes = self.definition.scalar.input_dtypes
        transfer_dtype = self.definition.scalar.numpy_dtype()
        shape = tuple(list(self.shape()) + [self.definition.length()])

        if not isinstance(array, np.ndarray):
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(type(array), self.glsl_dtype(), np.ndarray, self.path_to_str(path)))
        if array.dtype.type not in numpy_dtypes:
            raise TypeError("Got datatype {} for {} variable, expected {} at {}"
                            .format(array.dtype, self.glsl_dtype(), numpy_dtypes, self.path_to_str(path)))
        if tuple(array.shape) != shape:
            raise TypeError("Got shape {} for {} variable, expected {} at {}"
                            .format(array.shape, self.glsl_dtype(), shape, self.path_to_str(path)))

        p = (self.a - self.definition.size()) // self.definition.scalar.size()
        a = self.a // self.definition.scalar.size()

        array_padded = np.zeros(a * np.product(shape[:-1]), dtype=transfer_dtype)
        mask = (np.arange(len(array_padded)) % a) < (a - p)
        array_padded[mask] = array.flatten()

        return array_padded.tobytes()

    def to_bytes_for_matrices(self, array, path):
        numpy_dtype = self.definition.vector.scalar.numpy_dtype()
        shape = tuple(list(self.shape()) + list(self.definition.shape()))

        if not isinstance(array, np.ndarray):
            raise RuntimeError("Got datatype {} for {} variable, expected {} at {}"
                               .format(type(array), self.glsl_dtype(), np.ndarray, self.path_to_str(path)))
        if array.dtype != numpy_dtype:
            raise RuntimeError("Got datatype {} for {} variable, expected {} at {}"
                               .format(array.dtype, self.glsl_dtype(), numpy_dtype, self.path_to_str(path)))
        if tuple(array.shape) != shape:
            raise RuntimeError("Got shape {} for {} variable, expected {} at {}"
                               .format(array.shape, self.glsl_dtype(), shape, self.path_to_str(path)))

        # swap the last two dimensions if necessary
        if self.definition.order == Order.COLUMN_MAJOR:
            array = np.swapaxes(array, -2, -1)
            shape = array.shape

        p = (self.a - self.definition.vector.size()) // self.definition.vector.scalar.size()
        a = self.a // self.definition.vector.scalar.size()

        array_padded = np.zeros(a * np.product(shape[:-1]), dtype=array.dtype)
        mask = (np.arange(len(array_padded)) % a) < (a - p)
        array_padded[mask] = array.flatten()

        return array_padded.tobytes()

    def from_bytes(self, bytez):
        if isinstance(self.definition, Scalar):
            return self.from_bytes_for_scalars(bytez)

        elif isinstance(self.definition, Vector):
            return self.from_bytes_for_vectors(bytez)

        elif isinstance(self.definition, Matrix):
            return self.from_bytes_for_matrices(bytez)

        else:
            values = np.zeros(self.shape()).tolist()
            offset = 0
            size = self.definition.size()

            for indices in NdArray.iterate(self.shape()):
                NdArray.assign(values, indices, self.definition.from_bytes(bytez[offset:offset + size]))
                offset += size
                offset += (self.a - offset % self.a) % self.a  # bytes

            return values

    def from_bytes_for_scalars(self, bytez):
        if self.layout == Layout.STD430:
            array_flat = np.frombuffer(bytez, dtype=self.definition.numpy_dtype())

        else:
            p = (self.a - self.definition.alignment()) // self.definition.size()
            a = self.a // self.definition.size()

            array_padded = np.frombuffer(bytez, dtype=self.definition.numpy_dtype())
            mask = (np.arange(a * np.product(self.shape())) % a) < (a - p)
            array_flat = array_padded[mask]

        return array_flat.reshape(self.shape())

    def from_bytes_for_vectors(self, bytez):
        shape = tuple(list(self.shape()) + [self.definition.length()])
        p = (self.a - self.definition.size()) // self.definition.scalar.size()
        a = self.a // self.definition.scalar.size()

        array_padded = np.frombuffer(bytez, dtype=self.definition.scalar.numpy_dtype())
        mask = (np.arange(a * np.product(shape[:-1])) % a) < (a - p)
        return array_padded[mask].reshape(shape)

    def from_bytes_for_matrices(self, bytez):
        shape = tuple(list(self.shape()) + list(self.definition.shape()))
        p = (self.a - self.definition.vector.size()) // self.definition.vector.scalar.size()
        a = self.a // self.definition.vector.scalar.size()

        if self.definition.order == Order.COLUMN_MAJOR:
            shape = tuple(list(shape[:-2]) + [shape[-1], shape[-2]])

        array_padded = np.frombuffer(bytez, dtype=self.definition.vector.scalar.numpy_dtype())
        mask = (np.arange(a * np.product(shape[:-1])) % a) < (a - p)
        array = array_padded[mask].reshape(shape)

        if self.definition.order == Order.COLUMN_MAJOR:
            array = np.swapaxes(array, -2, -1)

        return array


class Struct(ByteRepresentation):

    def __init__(self, definitions, layout, member_names=None, type_name=None):
        super(Struct, self).__init__()
        self.definitions = definitions
        self.member_names = member_names or ([None] * len(definitions))
        self.type_name = type_name
        self.layout = layout  # precomputes alignment

        if len(set(definitions)) != len(definitions):
            raise BytesError("For struct members a definition object can not be used more than once")

    @property
    def layout(self):
        return self._layout

    @layout.setter
    def layout(self, layout):
        self._layout = layout

        # set children first
        for definition in self.definitions:
            if isinstance(definition, (Array, Matrix, Struct)):
                definition.layout = layout

        self.a = None
        self.precompute_alignment()

    def precompute_alignment(self):
        if self.layout in [Layout.STD140, Layout.STD430, Layout.STDXXX]:
            self.a = max([d.alignment() for d in self.definitions])

            if self.layout == Layout.STD140:
                self.a += (16 - self.a % 16) % 16

    def copy(self):
        return Struct([d.copy() for d in self.definitions], self.layout, self.member_names, self.type_name)

    def size(self):
        return self.steps()[-1]

    def offsets(self):
        return self.steps()[:-1]

    def steps(self):
        steps = [0]

        for d in self.definitions:
            a = d.alignment()
            padding_before = (a - steps[-1] % a) % a
            steps[-1] = steps[-1] + padding_before  # update last step + size to next step
            steps.append(steps[-1] + d.size())

        a = self.alignment()
        padding_after = (a - steps[-1] % a) % a
        steps[-1] = steps[-1] + padding_after

        return steps

    def alignment(self):
        return self.a

    def __str__(self, name=None, indent=2):
        # s = "struct [{}] {{\n".format(name or "?")
        s = "{} [{}] {{\n".format(self.type_name or "<struct_type>", name or "?")
        for i, (definition, member_name) in enumerate(zip(self.definitions, self.member_names)):
            s += "{}({}) {}\n".format(" " * indent, i, definition.__str__(name=member_name, indent=indent + 2))
        s += "{}}}".format(" " * (indent - 2))
        return s

    def glsl_dtype(self):
        if self.type_name is None:
            raise BytesError("Type name was not defined for structure")
        return self.type_name

    def __extend_path(self, path, member_index):
        if self.member_names[member_index]:
            step = "'{}'".format(self.member_names[member_index])
        else:
            step = str(member_index)
        return list(path) + ["member " + step]

    def compare(self, other, path=(), quiet=True):
        if not self.compare_type(type(self), type(other), path, quiet):
            return False
        if not self.compare_layout(self.layout, other.layout, path, quiet):
            return False

        definitions_expected = self.definitions
        definitions_other = other.definitions

        if len(definitions_expected) != len(definitions_other):
            if not quiet:
                raise TypeError("Expected {} members, but got {} members at {}".format(
                    len(definitions_expected), len(definitions_other), self.path_to_str(path)
                ))
            return False

        for i, (definition_expected, definition_other) in enumerate(zip(definitions_expected, definitions_other)):
            if not definition_expected.compare(definition_other, self.__extend_path(path, i), quiet):
                return False

        return True

    def to_bytes(self, values, path=()):
        bytez = bytearray()

        for i, d in enumerate(self.definitions):
            a = d.alignment()
            padding_before = (a - len(bytez) % a) % a
            bytez += bytearray(padding_before)
            bytez += d.to_bytes(values[d], self.__extend_path(path, i))

        # padding at the end
        a = self.alignment()
        padding_after = (a - len(bytez) % a) % a
        bytez += bytearray(padding_after)

        return bytez

    def from_bytes(self, bytez):
        values = {}
        offset = 0

        for i, d in enumerate(self.definitions):
            a = d.alignment()
            offset += (a - offset % a) % a  # padding before
            size = d.size()
            values[d] = d.from_bytes(bytez[offset:offset + size])
            offset += size

        # nothing needs to be done about the padding at the end
        return values


class ByteCache(object):

    def __init__(self, definition):
        if type(definition) != Struct:
            raise BytesError("ByteCaches can only be initialized with struct definitions")
        self.definition = definition
        self.values = {}
        self.dirty = False

        for i, d in enumerate(self.definition.definitions):
            value = None

            if isinstance(d, Struct):
                value = ByteCache(d)

            if Array.is_array_of_structs(d):
                value = np.zeros(d.shape()).tolist()

                for indices in NdArray.iterate(d.shape()):
                    NdArray.assign(value, indices, ByteCache(d.definition))

            self.values[d] = value

    def get_as_dict(self):
        data = {}

        for d in self.definition.definitions:
            value = self.values[d]

            if isinstance(d, Struct):
                value = self.values[d].get_as_dict()

            if Array.is_array_of_structs(d):
                value = np.zeros(d.shape()).tolist()

                for indices in NdArray.iterate(d.shape()):
                    cache = NdArray.get(self.values[d], indices)
                    NdArray.assign(value, indices, cache.get_as_dict())

            data[d] = value

        return data

    def set_from_dict(self, values):
        for d in self.definition.definitions:
            if isinstance(d, Struct):
                self.values[d].set_from_dict(values[d])

            elif isinstance(d, Array) and isinstance(d.definition, Struct):
                for indices in NdArray.iterate(d.shape()):
                    value = NdArray.get(values[d], indices)
                    cache = NdArray.get(self.values[d], indices)
                    cache.set_from_dict(value)

            else:
                self.values[d] = values[d]

    def set_dirty(self, dirty, include_children=True):
        self.dirty = dirty

        if include_children:
            for d in self.definition.definitions:
                value = self.values[d]

                if isinstance(d, Struct):
                    value.set_dirty(dirty, include_children)

                if Array.is_array_of_structs(d):
                    for indices in NdArray.iterate(d.shape()):
                        NdArray.get(value, indices).set_dirty(dirty, include_children)

    def is_dirty(self, include_children=True):
        if not include_children:
            return self.dirty

        dirty = self.dirty

        for d in self.definition.definitions:
            if dirty:
                return True

            value = self.values[d]

            if isinstance(d, Struct):
                dirty = dirty or value.is_dirty(include_children)

            if Array.is_array_of_structs(d):
                for indices in NdArray.iterate(d.shape()):
                    dirty = dirty or NdArray.get(value, indices).is_dirty(include_children)

        return dirty

    def __str__(self):
        s = self.__class__.__name__ + " around\n"
        s += str(self.definition)
        return s

    def __definition_from_key(self, key):
        if isinstance(key, int):
            return self.definition.definitions[key]

        if isinstance(key, str):
            if key not in self.definition.member_names:
                possible_reasons = [
                    "the definition was defined manually and the member names were not specified",
                    "the definition was taken from shader bytecode, but it did not contain the member names "
                    "(they are optional, the compiler must not put them in)"
                ]
                raise ValueError(
                    "Did not find {} in member names. In case this is not a typo please reference by index, member "
                    "names can be empty in following scenarios:\n{}".format(
                        key, "".join(["* {}\n".format(reason) for reason in possible_reasons])
                    )
                )
            index = self.definition.member_names.index(key)
            return self.definition.definitions[index]

        if isinstance(key, ByteRepresentation):
            return key

        raise ValueError("Invalid key")

    def __getitem__(self, key):
        return self.values[self.__definition_from_key(key)]

    def __setitem__(self, key, value):
        self.values[self.__definition_from_key(key)] = value
        self.dirty = True


# specs
# https://www.khronos.org/registry/vulkan/specs/1.1/html/chap14.html#interfaces-resources-layout
# https://github.com/KhronosGroup/glslang/issues/201#issuecomment-204785552 (example)
#
# https://www.khronos.org/registry/OpenGL/extensions/ARB/ARB_uniform_buffer_object.txt
#   CTRL+F "Sub-section 2.15.3.1.2"
#   CTRL+F "The following example illustrates the rules specified by the "std140" layout"
# https://www.khronos.org/registry/OpenGL/extensions/ARB/ARB_shader_storage_buffer_object.txt
# https://www.khronos.org/registry/OpenGL/extensions/ARB/ARB_arrays_of_arrays.txt
#
# Wiki layout bindings etc
# https://www.khronos.org/opengl/wiki/Interface_Block_(GLSL)
# https://www.khronos.org/opengl/wiki/Data_Type_(GLSL)
#
# more stuff to come
# http://www.paranormal-entertainment.com/idr/blog/posts/2014-01-29T17:08:42Z-GLSL_multidimensional_array_madness/



# -*- coding: UTF-8 -*-

import itertools
import struct

import lava.api.constants.spirv as spirv
from lava.api.util import LavaError


class ByteCodeError(LavaError):

    UNEXPECTED = "Something unexpected happened"

    def __init__(self, message):
        super(ByteCodeError, self).__init__(message)

    @classmethod
    def unexpected(cls):
        return cls(cls.UNEXPECTED)


class ByteCode(object):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf

    def __init__(self, bytez):
        self.bytez = bytearray(bytez)

        self.header = ByteCodeHeader(self.bytez)
        self.instructions = []

        # parse instruction by instruction
        step = self.header.words() * spirv.WORD_BYTE_SIZE
        while step < len(bytez):
            instruction = ByteCodeInstruction(self.bytez[step:])
            self.instructions.append(instruction)
            step += instruction.words() * spirv.WORD_BYTE_SIZE

        # recover types
        self.types_scalar, self.types_vector, self.types_matrix = self.find_basic_types()
        self.types_array = self.find_array_types()
        self.types_struct = self.find_struct_types()

    @classmethod
    def read_word(cls, bytez, offset=0):
        # word to unsigned integer
        return cls.read_words(bytez, n=1, offset=offset)[0]

    @classmethod
    def read_words(cls, bytez, n=-1, offset=0):
        # words to unsigned integers
        if n == -1:
            n = len(bytez) // spirv.WORD_BYTE_SIZE - offset
        a = offset * spirv.WORD_BYTE_SIZE
        b = a + n * spirv.WORD_BYTE_SIZE
        return struct.unpack("I" * n, bytez[a:b])

    @classmethod
    def read_words_as_string(cls, bytez, n=-1, offset=0):
        a = offset * spirv.WORD_BYTE_SIZE
        b = len(bytez) if n == -1 else a + n * spirv.WORD_BYTE_SIZE
        return bytez[a:b].rstrip(b"\0").decode("utf-8")

    def __str__(self):
        strings = []
        for instruction in self.instructions:
            if instruction.op is None:
                strings.append("Op id={} words={}".format(instruction.op_id, instruction.words()))
            else:
                strings.append(str(instruction.op))
        return "\n".join(strings)

    def abort(self):
        raise ByteCodeError.unexpected()

    def find_instructions(self, operation):
        results = []
        for instruction in self.instructions:
            if instruction.op_id == operation.ID:
                results.append(instruction)
        return results

    def find_instructions_with_attributes(self, operation, **attributes):
        results = self.find_instructions(operation)
        results_filtered = []

        for instruction in results:
            matches = 0

            for attr_key, attr_value in attributes.items():
                if attr_key not in instruction.op.__dict__:
                    break
                if instruction.op.__dict__[attr_key] != attr_value:
                    break
                matches += 1

            if matches == len(attributes):
                results_filtered.append(instruction)

        return results_filtered

    def find_basic_types(self):
        types_scalar = {}
        types_vector = {}
        types_matrix = {}

        search_scalar = [
            (spirv.DataType.FLOAT, {"operation": OpTypeFloat, "width": 32}),
            (spirv.DataType.DOUBLE, {"operation": OpTypeFloat, "width": 64}),
            (spirv.DataType.INT, {"operation": OpTypeInt, "width": 32, "signedness": 1}),
            (spirv.DataType.UINT, {"operation": OpTypeInt, "width": 32, "signedness": 0}),
        ]

        # scalar types are "standalone"
        for type_scalar, search_data in search_scalar:
            instructions = self.find_instructions_with_attributes(**search_data)
            if len(instructions) == 1:
                types_scalar[instructions[0].op.result_id] = type_scalar
            elif len(instructions) > 1:
                self.abort()

        # vector types reference the scalar types
        for result_id, n in itertools.product(types_scalar.keys(), range(2, 5)):
            instructions = self.find_instructions_with_attributes(operation=OpTypeVector, component_type=result_id,
                                                                  component_count=n)
            if len(instructions) == 1:
                types_vector[instructions[0].op.result_id] = (types_scalar[result_id], n)
            elif len(instructions) > 1:
                self.abort()

        # matrix types reference the vector types
        for result_id, cols in itertools.product(types_vector.keys(), range(2, 5)):
            instructions = self.find_instructions_with_attributes(operation=OpTypeMatrix, column_type=result_id,
                                                                  column_count=cols)
            if len(instructions) == 1:
                scalar_type, rows = types_vector[result_id]
                types_matrix[instructions[0].op.result_id] = (scalar_type, rows, cols)
            elif len(instructions) > 1:
                self.abort()

        return types_scalar, types_vector, types_matrix

    def find_array_types(self):
        types_array = {}
        tmp = {}

        # find all array types first
        for instruction in self.find_instructions(OpTypeArray):
            constants = self.find_instructions_with_attributes(OpConstant, result_id=instruction.op.length)
            if len(constants) == 1:
                n = constants[0].op.literals[0]
                tmp[instruction.op.result_id] = (instruction.op.element_type, [n])

        # collapse them into nd-arrays
        for idx in sorted(tmp.keys(), reverse=True):
            if idx in tmp:
                ref, dims = tmp[idx]

                while ref in tmp:
                    other_ref, other_dims = tmp[ref]
                    del tmp[ref]
                    ref = other_ref
                    dims += other_dims

                types_array[idx] = (ref, tuple(dims))
                del tmp[idx]

        return types_array

    def find_struct_types(self):
        types_struct = {}

        for instruction in self.find_instructions(OpTypeStruct):
            types_struct[instruction.op.result_id] = instruction.op.member_types

        return types_struct

    def find_names(self, struct_id):
        struct_name = None

        instructions = self.find_instructions_with_attributes(OpName, target_id=struct_id)
        if len(instructions) == 1:
            struct_name = instructions[0].op.name

        instructions = self.find_instructions_with_attributes(OpMemberName, type_id=struct_id)
        member_names = {instruction.op.member: instruction.op.name for instruction in instructions}
        return struct_name, [member_names.get(i) for i in range(max(member_names.keys()) + 1)]

    def find_member_ids(self, struct_id):
        instructions = self.find_instructions_with_attributes(OpTypeStruct, result_id=struct_id)
        if len(instructions) == 1:
            return instructions[0].op.member_types
        else:
            return None

    def find_offsets(self, struct_id):
        offsets = {}

        for instruction in self.find_instructions_with_attributes(OpMemberDecorate, decoration=spirv.Decoration.OFFSET,
                                                                  type_id=struct_id):
            member = instruction.op.member
            offset = instruction.op.literals[0]
            offsets[member] = offset

        return offsets

    def find_accesses(self, struct_id):
        accesses = {}

        instructions1 = self.find_instructions_with_attributes(OpMemberDecorate, type_id=struct_id,
                                                               decoration=spirv.Decoration.NON_READABLE)
        instructions2 = self.find_instructions_with_attributes(OpMemberDecorate, type_id=struct_id,
                                                               decoration=spirv.Decoration.NON_WRITABLE)

        for instruction in instructions1 + instructions2:
            member = instruction.op.member

            tmp = accesses.get(member, set())
            tmp.add(instruction.op.decoration)
            accesses[member] = tmp

        return accesses

    def find_strides(self, array_id):
        array_ids = [array_id]

        instructions = self.find_instructions_with_attributes(OpTypeArray, result_id=array_ids[-1])
        while len(instructions) == 1:
            array_ids.append(instructions[0].op.element_type)
            instructions = self.find_instructions_with_attributes(OpTypeArray, result_id=array_ids[-1])
        array_ids = array_ids[:-1]  # last type is no array

        strides = []

        for array_id in array_ids:
            instructions = self.find_instructions_with_attributes(OpDecorate, decoration=spirv.Decoration.ARRAY_STRIDE,
                                                                  target_id=array_id)
            if len(instructions) == 1:
                strides.append(instructions[0].op.literals[0])
            else:
                self.abort()

        return strides

    def find_matrix_stride(self, struct_id, member):
        instructions = self.find_instructions_with_attributes(OpMemberDecorate, type_id=struct_id, member=member,
                                                              decoration=spirv.Decoration.MATRIX_STRIDE)
        if len(instructions) == 1:
            return instructions[0].op.literals[0]
        else:
            return None

    def find_orders(self, struct_id):
        orders = {}

        instructions_row_major = self.find_instructions_with_attributes(OpMemberDecorate, type_id=struct_id,
                                                                        decoration=spirv.Decoration.ROW_MAJOR)

        instructions_col_major = self.find_instructions_with_attributes(OpMemberDecorate, type_id=struct_id,
                                                                        decoration=spirv.Decoration.COL_MAJOR)

        for instruction in instructions_row_major + instructions_col_major:
            member = instruction.op.member
            orders[member] = instruction.op.decoration

        return orders

    def find_blocks(self):
        blocks = {}
        blocks1 = self.find_instructions_with_attributes(OpDecorate, decoration=spirv.Decoration.BLOCK)  # ubo
        blocks2 = self.find_instructions_with_attributes(OpDecorate, decoration=spirv.Decoration.BUFFER_BLOCK)  # ssbo

        for candidate in blocks1 + blocks2:
            # find associated pointer
            instructions = self.find_instructions_with_attributes(OpTypePointer, type_id=candidate.op.target_id)
            if len(instructions) != 1:
                self.abort()

            # find associated variable
            instructions = self.find_instructions_with_attributes(OpVariable, result_type=instructions[0].op.result_id)
            if len(instructions) != 1:
                self.abort()
            storage_class = instructions[0].op.storage_class

            # find associated binding
            instructions = self.find_instructions_with_attributes(OpDecorate, target_id=instructions[0].op.result_id,
                                                                  decoration=spirv.Decoration.BINDING)

            block_type = candidate.op.decoration

            if len(instructions) == 1:
                binding_id = instructions[0].op.literals[0]
                blocks[candidate.op.target_id] = (block_type, storage_class, binding_id)

            else:
                self.abort()

        return blocks

    def find_entry_points(self, execution_model):
        entry_points = {}
        instructions = self.find_instructions_with_attributes(OpEntryPoint, execution_model=execution_model)

        for instruction in instructions:
            entry_points[instruction.op.entry_point] = instruction.op.name

        return entry_points

    def find_entry_point_details(self, entry_point_index):
        execution_mode = None
        literals = None
        instructions = self.find_instructions_with_attributes(OpExecutionMode, entry_point=entry_point_index)

        if len(instructions) == 1:
            execution_mode = instructions[0].op.execution_mode
            literals = instructions[0].op.literals
        else:
            self.abort()

        return execution_mode, literals


class ByteCodeHeader(object):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#subsection.2.3

    def __init__(self, bytez):
        self.bytez = bytez
        magic_number, version, generator_magic_number, bound, _ = ByteCode.read_words(bytez, spirv.WORD_COUNT_HEADER)

        if magic_number != spirv.MAGIC_NUMBER:
            raise ByteCodeError("MagicNumber does not match SPIR-V specs")

        self.version_major = (version & 0x00FF0000) >> 16
        self.version_minor = (version & 0x0000FF00) >> 8
        self.generator_magic_number = generator_magic_number
        self.bound = bound

    def words(self):
        return spirv.WORD_COUNT_HEADER


class ByteCodeInstruction(object):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#subsection.2.3

    def __init__(self, bytez):
        first_word = ByteCode.read_word(bytez)
        self.word_count = first_word >> 16
        self.op_id = first_word & 0x0000FFFF
        self.bytez = bytez[:self.word_count * spirv.WORD_BYTE_SIZE]
        self.op = None

        if self.op_id in OPS_REGISTER:
            self.op = OPS_REGISTER[self.op_id](self.bytez[spirv.WORD_BYTE_SIZE:])

    def words(self):
        return self.word_count


class Op(object):
    ID = -1

    def __init__(self, bytez):
        self.bytez = bytez

    def describe(self):
        raise NotImplementedError()

    def __str__(self):
        return "{:<20}{}".format(self.__class__.__name__, self.describe() or "")


class OpName(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpName
    ID = 5

    def __init__(self, bytez):
        super(OpName, self).__init__(bytez)
        self.target_id = ByteCode.read_word(bytez)
        self.name = ByteCode.read_words_as_string(bytez, offset=1)

    def describe(self):
        return "target_id={} name={}".format(self.target_id, self.name)


class OpMemberName(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpMemberName
    ID = 6

    def __init__(self, bytez):
        super(OpMemberName, self).__init__(bytez)
        self.type_id = ByteCode.read_word(bytez)
        self.member = ByteCode.read_word(bytez, offset=1)
        self.name = ByteCode.read_words_as_string(bytez, offset=2)

    def describe(self):
        return "type_id={} member_id={} name={}".format(self.type_id, self.member, self.name)


class OpEntryPoint(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpEntryPoint
    ID = 15

    def __init__(self, bytez):
        super(OpEntryPoint, self).__init__(bytez)
        self.execution_model = spirv.ExecutionModel.from_spirv(ByteCode.read_word(bytez))
        self.entry_point = ByteCode.read_word(bytez, offset=1)
        self.name = ByteCode.read_words_as_string(bytez, n=1, offset=2)
        self.ids = ByteCode.read_words(bytez, offset=3)

    def describe(self):
        return "execution_model={} entry_point={} name={} ids={}".format(self.execution_model, self.entry_point,
                                                                         self.name, self.ids)


class OpExecutionMode(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpExecutionMode
    ID = 16

    def __init__(self, bytez):
        super(OpExecutionMode, self).__init__(bytez)
        self.entry_point = ByteCode.read_word(bytez)
        self.execution_mode = spirv.ExecutionMode.from_spirv(ByteCode.read_word(bytez, offset=1))
        self.literals = ByteCode.read_words(bytez, offset=2)

    def describe(self):
        return "entry_point={} execution_mode={} literals={}".format(self.entry_point, self.execution_mode,
                                                                     self.literals)


class OpDecorate(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpDecorate
    ID = 71

    def __init__(self, bytez):
        super(OpDecorate, self).__init__(bytez)
        self.target_id = ByteCode.read_word(bytez)
        self.decoration = spirv.Decoration.from_spirv(ByteCode.read_word(bytez, offset=1))
        self.literals = ByteCode.read_words(bytez, offset=2)

    def describe(self):
        return "target_id={} decoration={} literals={}".format(self.target_id, self.decoration, self.literals)


class OpMemberDecorate(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpMemberDecorate
    ID = 72

    def __init__(self, bytez):
        super(OpMemberDecorate, self).__init__(bytez)
        self.type_id = ByteCode.read_word(bytez)
        self.member = ByteCode.read_word(bytez, offset=1)
        self.decoration = spirv.Decoration.from_spirv(ByteCode.read_word(bytez, offset=2))
        self.literals = ByteCode.read_words(bytez, n=(len(bytez) // spirv.WORD_BYTE_SIZE - 3), offset=3)

    def describe(self):
        return "type_id={} member={} decoration={} literals={}".format(self.type_id, self.member, self.decoration,
                                                                       self.literals)


class OpTypeVoid(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeVoid
    ID = 19

    def __init__(self, bytez):
        super(OpTypeVoid, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)

    def describe(self):
        return "result_id={}".format(self.result_id)


class OpTypeBool(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeBool
    ID = 20

    def __init__(self, bytez):
        super(OpTypeBool, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)

    def describe(self):
        return "result_id={}".format(self.result_id)


class OpTypeInt(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeInt
    ID = 21

    def __init__(self, bytez):
        super(OpTypeInt, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.width = ByteCode.read_word(bytez, offset=1)
        self.signedness = ByteCode.read_word(bytez, offset=2)

    def describe(self):
        return "result_id={} width={} signedness={}".format(self.result_id, self.width, self.signedness)


class OpTypeFloat(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeFloat
    ID = 22

    def __init__(self, bytez):
        super(OpTypeFloat, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.width = ByteCode.read_word(bytez, offset=1)

    def describe(self):
        return "result_id={} width={}".format(self.result_id, self.width)


class OpTypeVector(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeVector
    ID = 23

    def __init__(self, bytez):
        super(OpTypeVector, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.component_type = ByteCode.read_word(bytez, offset=1)
        self.component_count = ByteCode.read_word(bytez, offset=2)

    def describe(self):
        return "result_id={} component_type={} component_count={}".format(self.result_id, self.component_type,
                                                                          self.component_count)


class OpTypeMatrix(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeMatrix
    ID = 24

    def __init__(self, bytez):
        super(OpTypeMatrix, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.column_type = ByteCode.read_word(bytez, offset=1)
        self.column_count = ByteCode.read_word(bytez, offset=2)

    def describe(self):
        return "result_id={} column_type={} column_count={}".format(self.result_id, self.column_type, self.column_count)


class OpTypeImage(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeImage
    ID = 25

    def __init__(self, bytez):
        super(OpTypeImage, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.sampled_type = ByteCode.read_word(bytez, offset=1)
        # TODO: read out remaining 5 tons of information

    def describe(self):
        return "result_id={} sampled_type={}".format(self.result_id, self.sampled_type)


class OpTypeSampler(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeSampler
    ID = 26

    def __init__(self, bytez):
        super(OpTypeSampler, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)

    def describe(self):
        return "result_id={}".format(self.result_id)


class OpTypeArray(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeArray
    ID = 28

    def __init__(self, bytez):
        super(OpTypeArray, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.element_type = ByteCode.read_word(bytez, offset=1)
        self.length = ByteCode.read_word(bytez, offset=2)

    def describe(self):
        return "result_id={} element_type={} length={}".format(self.result_id, self.element_type, self.length)


class OpTypeRuntimeArray(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeRuntimeArray
    ID = 29

    def __init__(self, bytez):
        super(OpTypeRuntimeArray, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.element_type = ByteCode.read_word(bytez, offset=1)

    def describe(self):
        return "result_id={} element_type={}".format(self.result_id, self.element_type)


class OpTypeStruct(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypeStruct
    ID = 30

    def __init__(self, bytez):
        super(OpTypeStruct, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.member_types = ByteCode.read_words(bytez, offset=1)

    def describe(self):
        return "result_id={} member_types={}".format(self.result_id, self.member_types)


class OpSource(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpSource
    ID = 3

    def __init__(self, bytez):
        super(OpSource, self).__init__(bytez)
        self.source_language = spirv.SourceLanguage.from_spirv(ByteCode.read_word(bytez))
        self.version = ByteCode.read_word(bytez, offset=1)
        # ignore other attributes

    def describe(self):
        return "source_language={} version={}".format(self.source_language, self.version)


class OpSourceExtension(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpSourceExtension
    ID = 4

    def __init__(self, bytez):
        super(OpSourceExtension, self).__init__(bytez)
        self.extension = ByteCode.read_words_as_string(bytez)

    def describe(self):
        return "extension={}".format(self.extension)


class OpTypePointer(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpTypePointer
    ID = 32

    def __init__(self, bytez):
        super(OpTypePointer, self).__init__(bytez)
        self.result_id = ByteCode.read_word(bytez)
        self.storage_class = spirv.StorageClass.from_spirv(ByteCode.read_word(bytez, offset=1))
        self.type_id = ByteCode.read_word(bytez, offset=2)

    def describe(self):
        return "result_id={} storage_class={} type_id={}".format(self.result_id, self.storage_class, self.type_id)


class OpVariable(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpVariable
    ID = 59

    def __init__(self, bytez):
        super(OpVariable, self).__init__(bytez)
        self.result_type = ByteCode.read_word(bytez)
        self.result_id = ByteCode.read_word(bytez, offset=1)
        self.storage_class = spirv.StorageClass.from_spirv(ByteCode.read_word(bytez, offset=2))
        # self.initializer = ... (optional)

    def describe(self):
        return "result_type={} result_id={} storage_class={}".format(self.result_type, self.result_id,
                                                                     self.storage_class)


class OpConstant(Op):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#OpConstant
    ID = 43

    def __init__(self, bytez):
        super(OpConstant, self).__init__(bytez)
        self.result_type = ByteCode.read_word(bytez)
        self.result_id = ByteCode.read_word(bytez, offset=1)
        self.literals = ByteCode.read_words(bytez, offset=2)
        # self.initializer = ... (optional)

    def describe(self):
        return "result_type={} result_id={} literals={}".format(self.result_type, self.result_id, self.literals)


OPS_REGISTER = {op.ID: op for op in [
    OpName, OpMemberName, OpEntryPoint, OpExecutionMode, OpDecorate, OpMemberDecorate, OpTypeBool, OpTypeInt,
    OpTypeFloat, OpTypeVector, OpTypeMatrix, OpTypeArray, OpTypeRuntimeArray, OpTypeStruct, OpTypeVoid, OpTypeImage,
    OpTypeSampler, OpSource, OpSourceExtension, OpTypePointer, OpVariable, OpConstant
]}



# -*- coding: UTF-8 -*-

import warnings

from future.utils import raise_with_traceback

from lava.api.constants.vk import BufferUsage
from lava.api.pipeline import ShaderOperation, Pipeline
from lava.api.shader import Shader as _Shader
from lava.api.util import Destroyable, LavaError

__all__ = ["Shader", "Stage"]


class Shader(Destroyable):

    def __init__(self, session, path, entry_point=None):
        super(Shader, self).__init__()
        self.session = session
        self.session.register_shader(self)
        self.vulkan_shader = _Shader(self.session.device, path, entry_point)
        self.vulkan_shader.inspect()

    def _destroy(self):
        self.vulkan_shader.destroy()

    def get_bindings(self):
        return self.vulkan_shader.get_bindings()

    def get_block_definition(self, binding):
        return self.vulkan_shader.get_block_definition(binding)

    def get_block_usage(self, binding):
        return self.vulkan_shader.get_block_usage(binding)

    def get_local_size(self):
        return self.vulkan_shader.get_local_size()

    def get_block_access(self, binding):
        return self.vulkan_shader.get_block_access(binding)


class Stage(Destroyable):

    def __init__(self, shader, bindings):
        super(Stage, self).__init__()
        self.session = shader.session
        self.session.register_stage(self)
        self.shader = shader
        self.bindings = bindings

        self.checked = False
        self.check_workgroups()
        memory_object_binding = self.check_block_definitions()
        self.checked = True

        self.pipeline = Pipeline(self.session.device, shader.vulkan_shader, memory_object_binding)
        self.operation = ShaderOperation(self.session.device, self.pipeline, self.session.queue_index)

    def _destroy(self):
        if self.checked:
            self.operation.destroy()
            self.pipeline.destroy()

    def check_workgroups(self):
        physical_device = self.session.device.physical_device

        x, y, z = self.shader.vulkan_shader.get_local_size()
        x_max, y_max, z_max = physical_device.get_maximum_work_group_sizes()
        group_invocations_max = physical_device.get_maximum_work_group_invocations()
        if x > x_max or y > y_max or z > z_max:
            msg = "Device supports work group sizes up to x={} y={} z={}, but shader defines x={} y={} z={}".format(
                x_max, y_max, z_max, x, y, z)
            raise LavaError(msg)

        if x * y * z > group_invocations_max:
            msg = "Device supports work group invocations up to {}, but shader defines {}*{}*{}={}".format(
                group_invocations_max, x, y, z, x*y*z)
            raise LavaError(msg)

    def check_block_definitions(self):
        memory_object_binding = {}
        bindings_shader = self.shader.get_bindings()
        max_uniform_size = self.session.device.physical_device.get_maximum_uniform_size()

        for binding, buffer in self.bindings.items():
            if binding not in bindings_shader:
                raise LavaError("Shader does not define binding {}".format(binding))

            usage_shader = self.shader.get_block_usage(binding)
            usage_buffer = buffer.get_block_usage()

            if usage_buffer != usage_shader:
                raise LavaError("Shader defines binding {} as {}, but got {}".format(
                    binding, usage_shader, usage_buffer))

            definition_shader = self.shader.get_block_definition(binding)
            definition_buffer = buffer.get_block_definition()

            try:
                definition_shader.compare(definition_buffer, quiet=False)
            except TypeError as e:
                msg = "Block definition mismatch of buffer and shader at binding {}".format(binding)
                msg += "\n" + e.args[0]
                raise_with_traceback(LavaError(msg))

            if usage_buffer == BufferUsage.UNIFORM_BUFFER:
                if definition_buffer.size() > max_uniform_size:
                    msg = "Uniform at binding {} will be larger than maximum of {}, weird things might happen".format(
                        binding, max_uniform_size)
                    warnings.warn(msg, UserWarning)

            memory_object_binding[binding] = buffer.get_vulkan_buffer()
        return memory_object_binding

    def record(self, x, y, z, after_stages=(),):
        x_max, y_max, z_max = self.session.device.physical_device.get_maximum_work_group_counts()
        if x > x_max or y > y_max or z > z_max:
            msg = "Device supports work group counts up to x={} y={} z={}, but requested are x={} y={} z={}".format(
                x_max, y_max, z_max, x, y, z)
            raise LavaError(msg)

        wait_events = [stage.executor.event for stage in after_stages]
        self.operation.record(x, y, z, wait_events)

    def run(self):
        for binding, buffer in self.bindings.items():
            access_modifier = self.shader.get_block_access(binding)
            buffer.before_stage(self, binding, access_modifier)

        self.operation.run()

    def wait(self):
        self.operation.wait()

        for binding, buffer in self.bindings.items():
            access_modifier = self.shader.get_block_access(binding)
            buffer.after_stage(self, binding, access_modifier)

    def run_and_wait(self):
        self.run()
        self.wait()


# class Flow(object):
#
#     def __init__(self, session):
#         pass

# buffers either have
#   unlimited reads
# or
#   one write and unlimited reads afterwards



# -*- coding: UTF-8 -*-

import ctypes

import vulkan as vk

from lava.api.constants.vk import DescriptorType, QueueType, TIMEOUT_FOREVER
from lava.api.util import Destroyable, Event, Fence


class Pipeline(Destroyable):

    def __init__(self, device, shader, memory_objects, push_constants=()):
        super(Pipeline, self).__init__()
        self.device = device
        self.shader = shader
        self.memory_objects = memory_objects
        self.descriptor_set_handle = None
        self.descriptor_pool_handle = None
        bindings = []

        for i, memory_obj in memory_objects.items():
            bindings.append(memory_obj.descriptor_set_layout(i))

        descriptor_layout_create_info = vk.VkDescriptorSetLayoutCreateInfo(flags=None, pBindings=bindings)
        self.descriptor_set_layout_handle = vk.vkCreateDescriptorSetLayout(self.device.handle,
                                                                           descriptor_layout_create_info, None)

        shader_stage_create_info = vk.VkPipelineShaderStageCreateInfo(stage=vk.VK_SHADER_STAGE_COMPUTE_BIT,
                                                                      module=shader.handle,
                                                                      pName=shader.get_entry_point().encode("ascii"))

        if len(push_constants) > 0:
            if len(push_constants) > 1:
                raise NotImplementedError("Using more than one push constant is not implemented")

            push_constant_range = vk.VkPushConstantRange(stageFlags=vk.VK_SHADER_STAGE_FRAGMENT_BIT, offset=0,
                                                         size=ctypes.sizeof(push_constants[0]))

            pipeline_layout_create_info = vk.VkPipelineLayoutCreateInfo(pSetLayouts=[self.descriptor_set_layout_handle],
                                                                        pPushConstantRanges=[push_constant_range])

        else:
            pipeline_layout_create_info = vk.VkPipelineLayoutCreateInfo(pSetLayouts=[self.descriptor_set_layout_handle])

        self.pipeline_layout_handle = vk.vkCreatePipelineLayout(self.device.handle, pipeline_layout_create_info, None)
        pipeline_create_info = vk.VkComputePipelineCreateInfo(stage=shader_stage_create_info,
                                                              layout=self.pipeline_layout_handle)

        self.handle = vk.vkCreateComputePipelines(self.device.handle, None, 1, [pipeline_create_info], None)

    def _destroy(self):
        if self.descriptor_pool_handle is not None:
            vk.vkDestroyDescriptorPool(self.device.handle, self.descriptor_pool_handle, None)
        vk.vkDestroyDescriptorSetLayout(self.device.handle, self.descriptor_set_layout_handle, None)
        vk.vkDestroyPipelineLayout(self.device.handle, self.pipeline_layout_handle, None)
        vk.vkDestroyPipeline(self.device.handle, self.handle, None)

    def allocate_descriptor_sets(self):
        pool_sizes = []
        descriptor_types = {DescriptorType.UNIFORM_BUFFER: 0, DescriptorType.STORAGE_BUFFER: 0}

        for memory_obj in self.memory_objects.values():
            descriptor_type = memory_obj.descriptor_type()
            descriptor_types[descriptor_type] += 1

        for descriptor_type, count in descriptor_types.items():
            if count > 0:
                pool_sizes.append(vk.VkDescriptorPoolSize(DescriptorType.to_vulkan(descriptor_type), count))

        pool_create_info = vk.VkDescriptorPoolCreateInfo(maxSets=1, pPoolSizes=pool_sizes)
        self.descriptor_pool_handle = vk.vkCreateDescriptorPool(self.device.handle, pool_create_info, None)

        allocate_info = vk.VkDescriptorSetAllocateInfo(descriptorPool=self.descriptor_pool_handle,
                                                       pSetLayouts=[self.descriptor_set_layout_handle])

        self.descriptor_set_handle = vk.vkAllocateDescriptorSets(self.device.handle, allocate_info)[0]

    def update_descriptor_sets(self):
        write_data = []

        for i, memory_obj in self.memory_objects.items():
            write_data.append(memory_obj.write_descriptor_set(self.descriptor_set_handle, i))

        vk.vkUpdateDescriptorSets(self.device.handle, len(write_data), write_data, 0, None)

    def get_memory_objects(self):
        return self.memory_objects

    def get_descriptor_set_handle(self):
        if self.descriptor_set_handle is None:
            raise RuntimeError("Descriptor set was not allocated yet")
        return self.descriptor_set_handle

    def get_pipeline_layout_handle(self):
        return self.pipeline_layout_handle


class ShaderOperation(Destroyable):

    def __init__(self, device, pipeline, queue_index):
        super(ShaderOperation, self).__init__()
        self.device = device
        self.pipeline = pipeline
        self.queue_index = queue_index
        self.event = Event(self.device)
        self.fence = Fence(self.device, signalled=False)

        self.pipeline.allocate_descriptor_sets()
        self.pipeline.update_descriptor_sets()

        # VK_COMMAND_POOL_CREATE_TRANSIENT_BIT
        command_pool_create_info = vk.VkCommandPoolCreateInfo(flags=vk.VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
                                                              queueFamilyIndex=queue_index)
        self.command_pool_handle = vk.vkCreateCommandPool(self.device.handle, command_pool_create_info, None)

        allocate_info = vk.VkCommandBufferAllocateInfo(commandPool=self.command_pool_handle,
                                                       level=vk.VK_COMMAND_BUFFER_LEVEL_PRIMARY, commandBufferCount=1)
        self.command_buffer_handle = vk.vkAllocateCommandBuffers(self.device.handle, allocate_info)[0]

    def _destroy(self):
        vk.vkDestroyCommandPool(self.device.handle, self.command_pool_handle, None)
        self.event.destroy()
        self.fence.destroy()

    def record(self, count_x, count_y, count_z, wait_events=()):
        vk.vkBeginCommandBuffer(self.command_buffer_handle,
                                vk.VkCommandBufferBeginInfo(flags=vk.VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT))

        vk.vkCmdBindPipeline(self.command_buffer_handle, vk.VK_PIPELINE_BIND_POINT_COMPUTE, self.pipeline.handle)

        vk.vkCmdBindDescriptorSets(self.command_buffer_handle, vk.VK_PIPELINE_BIND_POINT_COMPUTE,
                                   self.pipeline.get_pipeline_layout_handle(), firstSet=0, descriptorSetCount=1,
                                   pDescriptorSets=[self.pipeline.get_descriptor_set_handle()], dynamicOffsetCount=0,
                                   pDynamicOffsets=None)

        vk.vkCmdResetEvent(self.command_buffer_handle, self.event.handle, vk.VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT)

        if len(wait_events) > 0:
            # https://stackoverflow.com/questions/45680135/pipeline-barriers-across-multiple-shaders
            vk.vkCmdWaitEvents(self.command_buffer_handle, len(wait_events), [event.handle for event in wait_events],
                               srcStageMask=vk.VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                               dstStageMask=vk.VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
                               memoryBarrierCount=0, pMemoryBarriers=None, bufferMemoryBarrierCount=0,
                               pBufferMemoryBarriers=None, imageMemoryBarrierCount=0, pImageMemoryBarriers=None)

        # vk.vkCmdPushConstants()

        vk.vkCmdDispatch(self.command_buffer_handle, count_x, count_y, count_z)

        vk.vkCmdSetEvent(self.command_buffer_handle, self.event.handle, vk.VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT)

        vk.vkEndCommandBuffer(self.command_buffer_handle)

    def run(self):
        queue_handle = self.device.get_queue_handle(QueueType.COMPUTE)

        submit_info = vk.VkSubmitInfo(pCommandBuffers=[self.command_buffer_handle])

        vk.vkQueueSubmit(queue_handle, 1, [submit_info], self.fence.handle)

    def wait(self):
        vk.vkWaitForFences(self.device.handle, 1, [self.fence.handle], True, TIMEOUT_FOREVER)
        vk.vkResetFences(self.device.handle, 1, [self.fence.handle])

    def run_and_wait(self):
        self.run()
        self.wait()


class CopyOperation(Destroyable):

    def __init__(self, device, queue_index):
        super(CopyOperation, self).__init__()
        self.device = device
        self.queue_index = queue_index
        self.fence = Fence(self.device, signalled=False)

        command_pool_create_info = vk.VkCommandPoolCreateInfo(flags=vk.VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT,
                                                              queueFamilyIndex=queue_index)
        self.command_pool_handle = vk.vkCreateCommandPool(self.device.handle, command_pool_create_info, None)

        allocation_info = vk.VkCommandBufferAllocateInfo(commandPool=self.command_pool_handle, commandBufferCount=1,
                                                         level=vk.VK_COMMAND_BUFFER_LEVEL_PRIMARY)
        self.command_buffer_handle = vk.vkAllocateCommandBuffers(self.device.handle, allocation_info)[0]

    def _destroy(self):
        vk.vkDestroyCommandPool(self.device.handle, self.command_pool_handle, None)
        self.fence.destroy()

    def record(self, src_buffer, dst_buffer):
        vk.vkBeginCommandBuffer(self.command_buffer_handle,
                                vk.VkCommandBufferBeginInfo(flags=vk.VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT))

        region = vk.VkBufferCopy(0, 0, src_buffer.get_size())
        vk.vkCmdCopyBuffer(self.command_buffer_handle, src_buffer.handle, dst_buffer.handle, 1, [region])

        vk.vkEndCommandBuffer(self.command_buffer_handle)

    def run(self):
        queue_handle = self.device.get_queue_handle(QueueType.COMPUTE)

        submit_info = vk.VkSubmitInfo(pCommandBuffers=[self.command_buffer_handle])

        vk.vkQueueSubmit(queue_handle, 1, [submit_info], self.fence.handle)

    def wait(self):
        vk.vkWaitForFences(self.device.handle, 1, [self.fence.handle], True, TIMEOUT_FOREVER)
        vk.vkResetFences(self.device.handle, 1, [self.fence.handle])

    def run_and_wait(self):
        self.run()
        self.wait()



# -*- coding: UTF-8 -*-

import itertools

import vulkan as vk

from lava.api.constants.vk import DeviceType, MemoryType, QueueType, VALIDATION_LAYERS
from lava.api.memory import Memory
from lava.api.util import Destroyable


class PhysicalDevice(object):

    def __init__(self, handle):
        self.handle = handle

        # meta data
        properties = vk.vkGetPhysicalDeviceProperties(self.handle)
        self.device_name = properties.deviceName
        self.device_type = DeviceType.from_vulkan(properties.deviceType)
        self.uniform_max_bytes = properties.limits.maxUniformBufferRange
        self.work_group_max_invocations = properties.limits.maxComputeWorkGroupInvocations
        self.work_group_max_sizes = [properties.limits.maxComputeWorkGroupSize[i] for i in range(3)]
        self.work_group_max_counts = [properties.limits.maxComputeWorkGroupCount[i] for i in range(3)]

        # queue information
        queue_families = vk.vkGetPhysicalDeviceQueueFamilyProperties(self.handle)
        self.queue_support = {}
        self.queue_idx = {t: [] for t in QueueType.keys()}

        for idx, queue_family in enumerate(queue_families):
            for queue_type in QueueType.keys():
                support = bool(queue_family.queueFlags & QueueType.to_vulkan(queue_type))
                self.queue_support[queue_type] = self.queue_support.get(queue_type, False) or support
                if support:
                    self.queue_idx[queue_type].append(idx)

        # memory properties
        memory_properties = vk.vkGetPhysicalDeviceMemoryProperties(self.handle)
        self.memory_type_idx = {t: [] for t in MemoryType.keys()}
        self.memory_type_idx_to_heap = {}
        self.heap_sizes = {}

        for idx, memory_type in itertools.product(range(memory_properties.memoryTypeCount), MemoryType.keys()):
            support = memory_properties.memoryTypes[idx].propertyFlags & MemoryType.to_vulkan(memory_type)
            if support:
                self.memory_type_idx[memory_type].append(idx)

            heap_idx = memory_properties.memoryTypes[idx].heapIndex
            self.memory_type_idx_to_heap[idx] = heap_idx

            if heap_idx not in self.heap_sizes:
                self.heap_sizes[heap_idx] = memory_properties.memoryHeaps[heap_idx].size

    @classmethod
    def all(cls, instance):
        return [cls(handle) for handle in vk.vkEnumeratePhysicalDevices(instance.handle)]

    def __str__(self):
        return "\n".join(["{}: {}".format(key, self.__dict__[key]) for key in sorted(self.__dict__)])

    def get_name(self):
        return self.device_name

    def get_type(self):
        return self.device_type

    def get_queue_indices(self, queue_type):
        return self.queue_idx[queue_type]

    def supports_queue_type(self, queue_type):
        return self.queue_support[queue_type]

    def get_memory_index(self, memory_types, supported_memory_indices):
        sets_of_indices = []

        for memory_type in memory_types:
            sets_of_indices.append(set(self.memory_type_idx[memory_type]))

        indices = sets_of_indices[0]
        for other_indices in sets_of_indices[1:]:
            indices.intersection_update(other_indices)

        indices.intersection_update(set(supported_memory_indices))

        if len(indices) == 0:
            return None
        else:
            return min(list(indices))

    def get_maximum_uniform_size(self):
        return self.uniform_max_bytes

    def get_maximum_work_group_sizes(self):
        return self.work_group_max_sizes

    def get_maximum_work_group_counts(self):
        return self.work_group_max_counts

    def get_maximum_work_group_invocations(self):
        return self.work_group_max_invocations


class Device(Destroyable):

    def __init__(self, physical_device, queue_definitions, validation_lvl=None, extensions=()):
        super(Device, self).__init__()
        self.validation_lvl = validation_lvl
        self.physical_device = physical_device
        self.memories = []
        queue_create_infos = []

        for queue_idx in set(map(lambda type_and_idx: type_and_idx[1], queue_definitions)):
            queue_create_infos.append(vk.VkDeviceQueueCreateInfo(
                queueFamilyIndex=queue_idx,
                queueCount=1,
                pQueuePriorities=[1.0]
            ))

        features = vk.VkPhysicalDeviceFeatures()

        if validation_lvl:
            create_info = vk.VkDeviceCreateInfo(
                pQueueCreateInfos=queue_create_infos,
                ppEnabledLayerNames=VALIDATION_LAYERS,
                ppEnabledExtensionNames=extensions,
                pEnabledFeatures=features
            )
        else:
            create_info = vk.VkDeviceCreateInfo(
                pQueueCreateInfos=queue_create_infos,
                enabledLayerCount=0,
                ppEnabledExtensionNames=extensions,
                pEnabledFeatures=features,
            )

        self.handle = vk.vkCreateDevice(physical_device.handle, create_info, None)
        self.queue_handles = {}
        self.queue_indices = {}

        for queue_type, queue_idx in queue_definitions:
            self.queue_handles[queue_type] = vk.vkGetDeviceQueue(self.handle, queue_idx, 0)
            self.queue_indices[queue_type] = queue_idx

    def _destroy(self):
        for memory in self.memories:
            memory.destroy()
        vk.vkDestroyDevice(self.handle, None)  # does also destroy the queues

    def get_physical_device(self):
        return self.physical_device

    def get_queue_handle(self, queue_type):
        return self.queue_handles[queue_type]

    def get_queue_index(self, queue_type):
        return self.queue_indices[queue_type]

    def allocate_memory(self, memory_types, size, supported_memory_indices):
        memory_index = self.physical_device.get_memory_index(memory_types, supported_memory_indices)

        if memory_index is None:
            raise RuntimeError("Could not find a device memory which supports {}".format(" and ".join(memory_types)))

        memory = Memory(self, memory_index, size)
        self.memories.append(memory)
        return memory

    def free_memory(self, memory):
        pass



# -*- coding: UTF-8 -*-

from functools import reduce
import itertools
import logging
import operator

import vulkan as vk


class Destroyable(object):

    def __init__(self):
        self.__destroyed = False

    def __del__(self):
        if not self.__destroyed:
            self.destroy()

    def destroy(self):
        if not self.__destroyed:
            self._destroy()
            self.__destroyed = True

    def _destroy(self):
        raise NotImplementedError()


class LavaError(Exception):

    def __init__(self, message):
        super(LavaError, self).__init__(message)


class LavaUnsupportedError(LavaError):

    def __init__(self, message):
        super(LavaUnsupportedError, self).__init__(message)


class Debugger(Destroyable):

    def __init__(self, instance, lvl=logging.INFO):
        super(Debugger, self).__init__()
        self.instance = instance
        self.lvl = lvl
        self.handle = None
        self.history = []
        self.history_size = 100
        self.attach()

    def _destroy(self):
        self.detach()

    def log(self, flags, object_type, object, location, message_code, layer, message, user_data):
        lvl = logging.DEBUG

        if flags & vk.VK_DEBUG_REPORT_INFORMATION_BIT_EXT:
            lvl = logging.INFO
        if flags & vk.VK_DEBUG_REPORT_WARNING_BIT_EXT:
            lvl = logging.WARNING
        if flags & vk.VK_DEBUG_REPORT_ERROR_BIT_EXT:
            lvl = logging.ERROR

        message_str = message[::]
        self.history.insert(0, message_str)
        if len(self.history) > self.history_size:
            self.history = self.history[:self.history_size]

        print("[VULKAN] " + message_str)
        return 0

    def attach(self):
        all_bits = [vk.VK_DEBUG_REPORT_DEBUG_BIT_EXT, vk.VK_DEBUG_REPORT_INFORMATION_BIT_EXT,
                    vk.VK_DEBUG_REPORT_WARNING_BIT_EXT, vk.VK_DEBUG_REPORT_ERROR_BIT_EXT]
        all_lvls = [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR]

        create_info = vk.VkDebugReportCallbackCreateInfoEXT(
            flags=reduce(operator.or_, all_bits[all_lvls.index(self.lvl):]),
            pfnCallback=self.log
        )

        self.handle = self.instance.vkCreateDebugReportCallbackEXT(self.instance.handle, create_info, None)

    def detach(self):
        self.instance.vkDestroyDebugReportCallbackEXT(self.instance.handle, self.handle, None)


class Event(Destroyable):

    def __init__(self, device):
        super(Event, self).__init__()
        self.device = device
        self.handle = vk.vkCreateEvent(self.device.handle, vk.VkEventCreateInfo(), None)

    def _destroy(self):
        vk.vkDestroyEvent(self.device.handle, self.handle, None)


class Fence(Destroyable):

    def __init__(self, device, signalled):
        super(Fence, self).__init__()
        self.device = device

        if signalled:
            create_info = vk.VkFenceCreateInfo(flags=vk.VK_FENCE_CREATE_SIGNALED_BIT)
        else:
            create_info = vk.VkFenceCreateInfo()

        self.handle = vk.vkCreateFence(self.device.handle, create_info, None)

    def _destroy(self):
        vk.vkDestroyFence(self.device.handle, self.handle, None)


class NdArray(object):

    @classmethod
    def iterate(cls, dims):
        return itertools.product(*[range(d) for d in dims])

    @classmethod
    def assign(cls, nd_array, indices, value):
        data = nd_array

        for index in indices[:-1]:
            data = data[index]

        data[indices[-1]] = value

    @classmethod
    def get(cls, nd_array, indices):
        data = nd_array

        for index in indices:
            data = data[index]

        return data



# -*- coding: UTF-8 -*-

import atexit
import logging
import os
import platform
import warnings


__version__ = "0.3.0"

ENV_VAR_SDK = "VULKAN_SDK"
ENV_VAR_LAYER_PATH = "VK_LAYER_PATH"

VALIDATION_LEVEL_DEBUG = logging.DEBUG
VALIDATION_LEVEL_INFO = logging.INFO
VALIDATION_LEVEL_WARNING = logging.WARNING
VALIDATION_LEVEL_ERROR = logging.ERROR

VALIDATION_LEVEL = None


__instance = None
__instance_usages = 0
__devices = []


try:
    import vulkan as vk
    __error = None
except Exception as e:
    __error = e

if ENV_VAR_SDK not in os.environ:
    __error = ImportError("{} environment variable not found".format(ENV_VAR_SDK))


def __initialize():
    global __error, __instance, __instance_usages, __devices, VALIDATION_LEVEL

    if __error:
        return

    from .api.constants.vk import QueueType
    from .api.device import PhysicalDevice
    from .api.instance import Instance

    if VALIDATION_LEVEL is not None and platform.system() == "Linux":
        if ENV_VAR_LAYER_PATH not in os.environ:
            __error = ImportError("{} environment variable not found (required for validations)"
                                  .format(ENV_VAR_LAYER_PATH))
            return

    try:
        __instance = Instance(validation_lvl=VALIDATION_LEVEL)
        __instance_usages = 0
        __devices = []

        for candidate in PhysicalDevice.all(__instance):
            if candidate.supports_queue_type(QueueType.COMPUTE):
                __devices.append(candidate)

        if len(__devices) == 0:
            warnings.warn("Did not find any suitable device", RuntimeWarning)

    except:
        __error = ImportError("Could not initialize {}".format(__name__))


def initialized():
    global __error
    return __error is None


def instance():
    global __error, __instance, __instance_usages, __devices, VALIDATION_LEVEL

    if __error:
        raise __error

    if __instance.validation_lvl != VALIDATION_LEVEL:
        if __instance_usages > 0:
            __instance_usages = 0
            warnings.warn("Recreating Vulkan instance with new validation level, any previously created sessions, "
                          "devices, etc. will be no longer usable", UserWarning)

        __cleanup()
        __initialize()

    __instance_usages += 1
    return __instance


def devices():
    instance()  # validation level might has been changed
    return __devices


@atexit.register
def __cleanup():
    global __instance, __devices

    if __instance is not None:
        from .session import sessions
        for sess in sessions:
            sess.destroy()

        del __devices
        __instance.destroy()


__initialize()

if __error is None:
    from .buffer import *
    from .session import *
    from .shader import *
    from .util import *



# -*- coding: UTF-8 -*-

import vulkan as vk

from lava.api.constants.vk import VALIDATION_LAYERS
from lava.api.util import Debugger, Destroyable


class Instance(Destroyable):

    def __init__(self, extensions=(), validation_lvl=None):
        super(Instance, self).__init__()
        self.validation_lvl = validation_lvl
        extensions = list(extensions)

        app_info = vk.VkApplicationInfo(
            pApplicationName=self.__class__.__name__,
            apiVersion=vk.VK_API_VERSION_1_0
        )

        if validation_lvl:
            create_info = vk.VkInstanceCreateInfo(
                pApplicationInfo=app_info,
                ppEnabledExtensionNames=extensions + [vk.VK_EXT_DEBUG_REPORT_EXTENSION_NAME],
                ppEnabledLayerNames=VALIDATION_LAYERS,
            )

        else:
            create_info = vk.VkInstanceCreateInfo(
                pApplicationInfo=app_info,
                ppEnabledExtensionNames=extensions,
                enabledLayerCount=0
            )

        self.handle = vk.vkCreateInstance(create_info, None)
        self.debugger = Debugger(self, lvl=validation_lvl) if validation_lvl else None

    def _destroy(self):
        if self.debugger:
            self.debugger.destroy()
        vk.vkDestroyInstance(self.handle, None)

    def __getattr__(self, item):
        # syntactic sugar to call vulkan instance functions as "pseudo methods" on this object, i.e.
        if item in vk._vulkan._instance_ext_funcs:
            def wrapper(*args, **kwargs):
                return vk.vkGetInstanceProcAddr(self.handle, item)(*args, **kwargs)
            return wrapper
        else:
            super(Instance, self).__getattribute__(item)



# -*- coding: UTF-8 -*-

import vulkan as vk


VALIDATION_LAYERS = ["VK_LAYER_LUNARG_standard_validation"]
SHADER_ENTRY = "main"
TIMEOUT_FOREVER = 0xFFFFFFFFFFFFFFFF


class VulkanEnum(object):

    _MAP = ()

    @classmethod
    def map(cls, key, keys, values):
        return values[keys.index(key)]

    @classmethod
    def to_vulkan(cls, key):
        keys, values = zip(*cls._MAP)
        return cls.map(key, keys, values)

    @classmethod
    def from_vulkan(cls, key):
        values, keys = zip(*cls._MAP)
        return cls.map(key, keys, values)

    @classmethod
    def keys(cls):
        return list(zip(*cls._MAP))[0]


class DeviceType(VulkanEnum):

    CPU = "CPU"
    DISCRETE_GPU = "DISCRETE_CPU"
    INTEGRATED_GPU = "INTEGRATED_GPU"
    VIRTUAL_GPU = "VIRTUAL_GPU"
    OTHER = "OTHER"

    _MAP = (
        (CPU, vk.VK_PHYSICAL_DEVICE_TYPE_CPU),
        (DISCRETE_GPU, vk.VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU),
        (INTEGRATED_GPU, vk.VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU),
        (VIRTUAL_GPU, vk.VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU)
    )


class QueueType(VulkanEnum):

    GRAPHICS = "GRAPHICS"
    COMPUTE = "COMPUTE"
    TRANSFER = "TRANSFER"
    SPARSE_BINDING = "SPARSE_BINDING"

    _MAP = (
        (GRAPHICS, vk.VK_QUEUE_GRAPHICS_BIT),
        (COMPUTE, vk.VK_QUEUE_COMPUTE_BIT),
        (TRANSFER, vk.VK_QUEUE_TRANSFER_BIT),
        (SPARSE_BINDING, vk.VK_QUEUE_SPARSE_BINDING_BIT),
    )


class MemoryType(VulkanEnum):
    # https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkMemoryPropertyFlagBits.html

    DEVICE_LOCAL = "DEVICE_LOCAL"
    HOST_VISIBLE = "HOST_VISIBLE"
    HOST_COHERENT = "HOST_COHERENT"
    HOST_CACHED = "HOST_CACHED"
    LAZILY_ALLOCATED = "LAZILY_ALLOCATED"
    # PROTECTED = "PROTECTED"

    CPU = [HOST_COHERENT, HOST_VISIBLE]
    GPU = [DEVICE_LOCAL]

    _MAP = (
        (DEVICE_LOCAL, vk.VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT),
        (HOST_VISIBLE, vk.VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT),
        (HOST_COHERENT, vk.VK_MEMORY_PROPERTY_HOST_COHERENT_BIT),
        (HOST_CACHED, vk.VK_MEMORY_PROPERTY_HOST_CACHED_BIT),
        (LAZILY_ALLOCATED, vk.VK_MEMORY_PROPERTY_LAZILY_ALLOCATED_BIT),
    )


class BufferUsage(VulkanEnum):
    # https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkBufferUsageFlagBits.html

    UNIFORM_BUFFER = "UNIFORM_BUFFER"
    STORAGE_BUFFER = "STORAGE_BUFFER"
    TRANSFER_SRC = "TRANSFER_SRC"
    TRANSFER_DST = "TRANSFER_DST"

    _MAP = (
        (UNIFORM_BUFFER, vk.VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT),
        (STORAGE_BUFFER, vk.VK_BUFFER_USAGE_STORAGE_BUFFER_BIT),
        (TRANSFER_SRC, vk.VK_BUFFER_USAGE_TRANSFER_SRC_BIT),
        (TRANSFER_DST, vk.VK_BUFFER_USAGE_TRANSFER_DST_BIT)
    )


class DescriptorType(VulkanEnum):
    # https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkDescriptorType.html

    SAMPLER = "SAMPLER"
    UNIFORM_BUFFER = "UNIFORM_BUFFER"
    STORAGE_BUFFER = "STORAGE_BUFFER"

    _MAP = (
        (SAMPLER, vk.VK_DESCRIPTOR_TYPE_SAMPLER),
        (UNIFORM_BUFFER, vk.VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER),
        (STORAGE_BUFFER, vk.VK_DESCRIPTOR_TYPE_STORAGE_BUFFER),
    )


class CommandBufferUsage(VulkanEnum):
    # https://www.khronos.org/registry/vulkan/specs/1.1-extensions/man/html/VkCommandBufferUsageFlagBits.html

    ONE_TIME_SUBMIT = "ONE_TIME_SUBMIT"
    SIMULTANEOUS_USE = "SIMULTANEOUS"

    _MAP = (
        (ONE_TIME_SUBMIT, vk.VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT),
        (SIMULTANEOUS_USE, vk.VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT)
    )







# -*- coding: UTF-8 -*-

import lava
from lava.api.constants.vk import QueueType
from lava.api.device import Device
from lava.api.util import Destroyable

__all__ = ["Session"]

sessions = set()


class Session(Destroyable):

    def __init__(self, physical_device, queue_index=None):
        super(Session, self).__init__()

        self.instance = lava.instance()  # validation level might has been changed
        if physical_device not in lava.devices():
            raise RuntimeError("Provided invalid / outdated device object")

        self.queue_index = queue_index or physical_device.get_queue_indices(QueueType.COMPUTE)[0]
        self.device = Device(physical_device, [(QueueType.COMPUTE, self.queue_index)],
                             validation_lvl=lava.VALIDATION_LEVEL)

        self.buffers = set()
        self.shaders = set()
        self.stages = set()

        sessions.add(self)

    def _destroy(self):
        for stage in self.stages:
            stage.destroy()
        for shader in self.shaders:
            shader.destroy()
        for buffer in self.buffers:
            buffer.destroy()
        self.device.destroy()

    def register_buffer(self, buffer):
        self.buffers.add(buffer)

    def register_shader(self, shader):
        self.shaders.add(shader)

    def register_stage(self, stage):
        self.stages.add(stage)



# -*- coding: UTF-8 -*-

WORD_BYTE_SIZE = 4
MAGIC_NUMBER = 0x07230203
WORD_COUNT_HEADER = 5


class DataType(object):
    # BOOL = "bool"  # machine unit 1 byte? (probably not 1 bit)
    INT = "int"
    UINT = "uint"
    FLOAT = "float"
    DOUBLE = "double"

    ALL = [INT, UINT, FLOAT, DOUBLE]


class Layout(object):
    STD140 = "std140"
    STD430 = "std430"
    STDXXX = "stdxxx"  # can not be inferred / does not matter


class Order(object):
    COLUMN_MAJOR = "column_major"
    ROW_MAJOR = "row_major"


class Access(object):
    READ_ONLY = "read_only"
    WRITE_ONLY = "write_only"
    READ_WRITE = "read_write"
    NEITHER = "neither"  # only "metadata" can be accessed, e.g. length of an array


class SpirvEnum(object):

    _MAP = ()
    _UNKNOWN = "UNKNOWN"

    @classmethod
    def map(cls, key, keys, values, default=None):
        if key in keys:
            return values[keys.index(key)]
        if default:
            return default
        raise RuntimeError("Could not map {}".format(key))

    @classmethod
    def to_spirv(cls, key):
        keys, values = zip(*cls._MAP)
        return cls.map(key, keys, values)

    @classmethod
    def from_spirv(cls, key):
        values, keys = zip(*cls._MAP)
        return cls.map(key, keys, values, default=cls._UNKNOWN)

    @classmethod
    def keys(cls):
        return list(zip(*cls._MAP))[0]


class SourceLanguage(SpirvEnum):

    UNKNOWN = "UNKNOWN"
    ESSL = "ESSL"
    GLSL = "GLSL"
    OPEN_CL_C = "OPEN_CL_C"
    OPEN_CL_CPP = "OPEN_CL_CPP"
    HLSL = "HLSL"

    _MAP = (
        (UNKNOWN, 0), (ESSL, 1), (GLSL, 2), (OPEN_CL_C, 3), (OPEN_CL_CPP, 4), (HLSL, 5),
    )


class ExecutionModel(SpirvEnum):

    VERTEX = "VERTEX"
    TESSELATION_CONTROL = "TESSELATION_CONTROL"
    TESSELATION_EVALUATION = "TESSELATION_EVALUATION"
    GEOMETRY = "GEOMETRY"
    FRAGMENT = "FRAGMENT"
    GL_COMPUTE = "GL_COMPUTE"
    KERNEL = "KERNEL"

    _MAP = (
        (VERTEX, 0), (TESSELATION_CONTROL, 1), (TESSELATION_EVALUATION, 2), (GEOMETRY, 3), (FRAGMENT, 4),
        (GL_COMPUTE, 5), (KERNEL, 6)
    )


class ExecutionMode(SpirvEnum):

    INVOCATIONS = "INVOCATIONS"
    SPACING_EQUAL = "SPACING_EQUAL"
    SPACING_FRACTIONAL_EVEN = "SPACING_FRACTIONAL_EVEN"
    SPACING_FRACTIONAL_ODD = "SPACING_FRACTIONAL_ODD"
    VERTEX_ORDER_CW = "VERTEX_ORDER_CW"
    VERTEX_ORDER_CCW = "VERTEX_ORDER_CCW"
    PIXEL_CENTER_INTEGER = "PIXEL_CENTER_INTEGER"
    ORIGIN_UPPER_LEFT = "ORIGIN_UPPER_LEFT"
    ORIGIN_LOWER_LEFT = "ORIGIN_LOWER_LEFT"
    EARLY_FRAGMENT_TESTS = "EARLY_FRAGMENT_TESTS"
    POINT_MODE = "POINT_MODE"
    XFB = "XFB"
    DEPTH_REPLACING = "DEPTH_REPLACING"
    DEPTH_GREATER = "DEPTH_GREATER"
    DEPTH_LESS = "DEPTH_LESS"
    DEPTH_UNCHANGED = "DEPTH_UNCHANGED"
    LOCAL_SIZE = "LOCAL_SIZE"
    LOCAL_SIZE_HINT = "LOCAL_SIZE_HINT"
    INPUT_POINTS = "INPUT_POINTS"
    INPUT_LINES = "INPUT_LINES"
    INPUT_LINES_ADJACENCY = "INPUT_LINES_ADJACENCY"
    TRIANGLES = "TRIANGLES"
    INPUT_TRIANGLES_ADJACENCY = "INPUT_TRIANGLES_ADJACENCY"
    QUADS = "QUADS"
    ISOLINES = "ISOLINES"
    OUTPUT_VERTICES = "OUTPUT_VERTICES"
    OUTPUT_POINTS = "OUTPUT_POINTS"
    OUTPUT_LINE_STRIP = "OUTPUT_LINE_STRIP"
    OUTPUT_TRIANGLE_STRIP = "OUTPUT_TRIANGLE_STRIP"
    VEC_TYPE_HINT = "VEC_TYPE_HINT"
    CONTRACTION_OFF = "CONTRACTION_OFF"
    INITIALIZER = "INITIALIZER"
    FINALIZER = "FINALIZER"
    SUBGROUP_SIZE = "SUBGROUP_SIZE"
    SUBGROUPS_PER_WORKGROUP = "SUBGROUPS_PER_WORKGROUP"
    SUBGROUPS_PER_WORKGROUP_ID = "SUBGROUPS_PER_WORKGROUP_ID"
    LOCAL_SIZE_ID = "LOCAL_SIZE_ID"
    LOCAL_SIZE_HINT_ID = "LOCAL_SIZE_HINT_ID"
    POST_DEPTH_COVERAGE = "POST_DEPTH_COVERAGE"
    STENCIL_REF_REPLACING_EXT = "STENCIL_REF_REPLACING_EXT"

    _MAP = (
        (INVOCATIONS, 0), (SPACING_EQUAL, 1), (SPACING_FRACTIONAL_EVEN, 2), (SPACING_FRACTIONAL_ODD, 3),
        (VERTEX_ORDER_CW, 4), (VERTEX_ORDER_CCW, 5), (PIXEL_CENTER_INTEGER, 6), (ORIGIN_UPPER_LEFT, 7),
        (ORIGIN_LOWER_LEFT, 8), (EARLY_FRAGMENT_TESTS, 9), (POINT_MODE, 10), (XFB, 11), (DEPTH_REPLACING, 12),
        (DEPTH_GREATER, 14), (DEPTH_LESS, 15), (DEPTH_UNCHANGED, 16), (LOCAL_SIZE, 17), (LOCAL_SIZE_HINT, 18),
        (INPUT_POINTS, 19), (INPUT_LINES, 20), (INPUT_LINES_ADJACENCY, 21), (TRIANGLES, 22),
        (INPUT_TRIANGLES_ADJACENCY, 23), (QUADS, 24), (ISOLINES, 25), (OUTPUT_VERTICES, 26), (OUTPUT_POINTS, 27),
        (OUTPUT_LINE_STRIP, 28), (OUTPUT_TRIANGLE_STRIP, 29), (VEC_TYPE_HINT, 30), (CONTRACTION_OFF, 31),
        (INITIALIZER, 33), (FINALIZER, 34), (SUBGROUP_SIZE, 35), (SUBGROUPS_PER_WORKGROUP, 36),
        (SUBGROUPS_PER_WORKGROUP_ID, 37), (LOCAL_SIZE_ID, 38), (LOCAL_SIZE_HINT_ID, 39),
        (POST_DEPTH_COVERAGE, 4446), (STENCIL_REF_REPLACING_EXT, 5027),
    )


class Decoration(SpirvEnum):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#Decoration

    RELAXED_PRECISION = "RELAXED_PRECISION"
    SPEC_ID = "SPEC_ID"
    BLOCK = "BLOCK"
    BUFFER_BLOCK = "BUFFER_BLOCK"
    ROW_MAJOR = "ROW_MAJOR"
    COL_MAJOR = "COL_MAJOR"
    ARRAY_STRIDE = "ARRAY_STRIDE"
    MATRIX_STRIDE = "MATRIX_STRIDE"
    GLSL_SHARED = "GLSL_SHARED"
    GLSL_PACKED = "GLSL_PACKED"
    C_PACKED = "C_PACKED"
    BUILT_IN = "BUILT_IN"
    NO_PERSPECTIVE = "NO_PERSPECTIVE"
    FLAT = "FLAT"
    PATCH = "PATCH"
    CENTROID = "CENTROID"
    SAMPLE = "SAMPLE"
    INVARIANT = "INVARIANT"
    RESTRICT = "RESTRICT"
    ALIASED = "ALIASED"
    VOLATILE = "VOLATILE"
    CONSTANT = "CONSTANT"
    COHERENT = "COHERENT"
    NON_WRITABLE = "NON_WRITABLE"
    NON_READABLE = "NON_READABLE"
    UNIFORM = "UNIFORM"
    SATURATED_CONVERSION = "SATURATED_CONVERSION"
    STREAM = "STREAM"
    LOCATION = "LOCATION"
    COMPONENT = "COMPONENT"
    INDEX = "INDEX"
    BINDING = "BINDING"
    DESCRIPTOR_SET = "DESCRIPTOR_SET"
    OFFSET = "OFFSET"
    XFB_BUFFER = "XFB_BUFFER"
    XFB_STRIDE = "XFB_STRIDE"
    FUNC_PARAM_ATTR = "FUNC_PARAM_ATTR"
    FP_ROUNDING_MODE = "FP_ROUNDING_MODE"
    FP_FAST_MATH_MODE = "FP_FAST_MATH_MODE"
    LINKAGE_ATTRIBUTES = "LINKAGE_ATTRIBUTES"
    NO_CONTRACTION = "NO_CONTRACTION"
    INPUT_ATTACHMENT_INDEX = "INPUT_ATTACHMENT_INDEX"
    ALIGNMENT = "ALIGNMENT"
    MAX_BYTE_OFFSET = "MAX_BYTE_OFFSET"
    ALIGNMENT_ID = "ALIGNMENT_ID"
    MAX_BYTE_OFFSET_ID = "MAX_BYTE_OFFSET_ID"
    EXPLICIT_INTERP_AMD = "EXPLICIT_INTERP_AMD"
    OVERRIDE_COVERAGE_NV = "OVERRIDE_COVERAGE_NV"
    PASSTHROUGH_NV = "PASSTHROUGH_NV"
    VIEWPORT_RELATIVE_NV = "VIEWPORT_RELATIVE_NV"
    SECONDARY_VIEWPORT_RELATIVE_NV = "SECONDARY_VIEWPORT_RELATIVE_NV"
    HLSL_COUNTER_BUFFER_GOOGLE = "HLSL_COUNTER_BUFFER_GOOGLE"
    HLSL_SEMANTIC_GOOGLE = "HLSL_SEMANTIC_GOOGLE"

    _MAP = (
        (RELAXED_PRECISION, 0), (SPEC_ID, 1), (BLOCK, 2), (BUFFER_BLOCK, 3), (ROW_MAJOR, 4), (COL_MAJOR, 5),
        (ARRAY_STRIDE, 6), (MATRIX_STRIDE, 7), (GLSL_SHARED, 8), (GLSL_PACKED, 9), (C_PACKED, 10), (BUILT_IN, 11),
        (NO_PERSPECTIVE, 13), (FLAT, 14), (PATCH, 15), (CENTROID, 16), (SAMPLE, 17), (INVARIANT, 18),
        (RESTRICT, 19), (ALIASED, 20), (VOLATILE, 21), (CONSTANT, 22), (COHERENT, 23), (NON_WRITABLE, 24),
        (NON_READABLE, 25), (UNIFORM, 26), (SATURATED_CONVERSION, 28), (STREAM, 29), (LOCATION, 30),
        (COMPONENT, 31), (INDEX, 32), (BINDING, 33), (DESCRIPTOR_SET, 34), (OFFSET, 35), (XFB_BUFFER, 36),
        (XFB_STRIDE, 37), (FUNC_PARAM_ATTR, 38), (FP_ROUNDING_MODE, 39), (FP_FAST_MATH_MODE, 40),
        (LINKAGE_ATTRIBUTES, 41), (NO_CONTRACTION, 42), (INPUT_ATTACHMENT_INDEX, 43), (ALIGNMENT, 44),
        (MAX_BYTE_OFFSET, 45), (ALIGNMENT_ID, 46), (MAX_BYTE_OFFSET_ID, 47), (EXPLICIT_INTERP_AMD, 4999),
        (OVERRIDE_COVERAGE_NV, 5248), (PASSTHROUGH_NV, 5250), (VIEWPORT_RELATIVE_NV, 5252),
        (SECONDARY_VIEWPORT_RELATIVE_NV, 5256), (HLSL_COUNTER_BUFFER_GOOGLE, 5634), (HLSL_SEMANTIC_GOOGLE, 5635),
    )


class StorageClass(SpirvEnum):
    # https://www.khronos.org/registry/spir-v/specs/1.2/SPIRV.pdf#subsection.3.7

    UNIFORM_CONSTANT = "UNIFORM_CONSTANT"
    INPUT = "INPUT"
    UNIFORM = "UNIFORM"
    OUTPUT = "OUTPUT"
    WORKGROUP = "WORKGROUP"
    CROSS_WORKGROUP = "CROSS_WORKGROUP"
    PRIVATE = "PRIVATE"
    FUNCTION = "FUNCTION"
    GENERIC = "GENERIC"
    PUSH_CONSTANT = "PUSH_CONSTANT"
    ATOMIC_COUNTER = "ATOMIC_COUNTER"
    IMAGE = "IMAGE"
    STORAGE_BUFFER = "STORAGE_BUFFER"

    _MAP = (
        (UNIFORM_CONSTANT, 0), (INPUT, 1), (UNIFORM, 2), (OUTPUT, 3), (WORKGROUP, 4), (CROSS_WORKGROUP, 5),
        (PRIVATE, 6), (FUNCTION, 7), (GENERIC, 8), (PUSH_CONSTANT, 9), (ATOMIC_COUNTER, 10), (IMAGE, 11),
        (STORAGE_BUFFER, 12),
    )



# -*- coding: UTF-8 -*-

import warnings

from lava.api.bytes import ByteCache, Struct
from lava.api.constants.spirv import Access
from lava.api.constants.vk import BufferUsage, MemoryType
from lava.api.memory import Buffer as _Buffer
from lava.api.pipeline import CopyOperation
from lava.api.util import Destroyable, LavaError, LavaUnsupportedError

__all__ = ["BufferCPU", "BufferGPU", "StagedBuffer"]


class BufferInterface(Destroyable):

    USAGE_STORAGE = BufferUsage.STORAGE_BUFFER
    USAGE_UNIFORM = BufferUsage.UNIFORM_BUFFER

    SYNC_EAGER = "eager"
    SYNC_LAZY = "lazy"
    SYNC_MANUAL = "manual"
    SYNC_DEFAULT = SYNC_LAZY

    def __init__(self, session, block_definition, block_usage):
        super(BufferInterface, self).__init__()
        if not isinstance(block_definition, Struct):
            raise LavaError("Block definitions must be structs")
        self.session = session
        self.block_definition = block_definition
        self.block_usage = block_usage

    def _destroy(self):
        raise NotImplementedError()

    def size(self):
        return self.block_definition.size()

    def get_block_definition(self):
        return self.block_definition

    def get_block_usage(self):
        return self.block_usage

    def get_vulkan_buffer(self):
        raise NotImplementedError()

    def before_stage(self, stage, binding, access_modifier):
        pass

    def after_stage(self, stage, binding, access_modifier):
        pass

    def flush(self):
        pass

    def fetch(self):
        pass

    def __getitem__(self, key):
        raise NotImplementedError()

    def __setitem__(self, key, value):
        raise NotImplementedError()


class Buffer(BufferInterface):

    LOCATION_CPU = "CPU"
    LOCATION_GPU = "GPU"

    def __init__(self, session, block_definition, block_usage, location):
        super(Buffer, self).__init__(session, block_definition, block_usage)
        self.location = location
        self.vulkan_buffer = None
        self.vulkan_memory = None
        self.session.register_buffer(self)
        self.allocate()
        self.copy_operation = CopyOperation(self.session.device, self.session.queue_index)

    def _destroy(self):
        self.vulkan_memory.destroy()
        self.vulkan_buffer.destroy()
        self.copy_operation.destroy()

    def allocate(self):
        if self.vulkan_buffer is not None:
            raise LavaError("Buffer is already allocated")

        self.vulkan_buffer = _Buffer(self.session.device, self.size(), self.block_usage, self.session.queue_index)

        minimum_size, _, supported_memory_indices = self.vulkan_buffer.get_memory_requirements()
        memory_types = {self.LOCATION_CPU: MemoryType.CPU, self.LOCATION_GPU: MemoryType.GPU}[self.location]

        self.vulkan_memory = self.session.device.allocate_memory(memory_types, minimum_size, supported_memory_indices)
        self.vulkan_buffer.bind_memory(self.vulkan_memory)

    def get_vulkan_buffer(self):
        return self.vulkan_buffer

    def get_location(self):
        return self.location

    def copy_to(self, other):
        self.copy_operation.record(self.vulkan_buffer, other.vulkan_buffer)
        self.copy_operation.run_and_wait()


class BufferCPU(Buffer):

    def __init__(self, session, block_definition, block_usage, sync_mode=BufferInterface.SYNC_DEFAULT):
        super(BufferCPU, self).__init__(session, block_definition, block_usage, Buffer.LOCATION_CPU)
        self.cache = ByteCache(self.block_definition)
        self.sync_mode = sync_mode
        self.fresh_bytez = False

    @classmethod
    def from_shader(cls, session, shader, binding, sync_mode=BufferInterface.SYNC_DEFAULT):
        block_definition = shader.get_block_definition(binding)
        block_usage = shader.get_block_usage(binding)
        return cls(session, block_definition, block_usage, sync_mode)

    def before_stage(self, stage, binding, access_modifier):
        if access_modifier in [Access.READ_ONLY, Access.READ_WRITE]:
            if self.sync_mode is self.SYNC_LAZY and self.cache.is_dirty():
                self.flush()

    def after_stage(self, stage, binding, access_modifier):
        if access_modifier in [Access.WRITE_ONLY, Access.READ_WRITE]:
            self.fresh_bytez = True
            if self.sync_mode is self.SYNC_EAGER:
                self.fetch()

    def __getitem__(self, key):
        if self.fresh_bytez and self.sync_mode is self.SYNC_LAZY:
            self.fetch()
        return self.cache[key]

    def __setitem__(self, key, value):
        self.cache[key] = value

        if self.sync_mode is self.SYNC_EAGER:
            self.flush()

    def is_synced(self):
        return not self.fresh_bytez and not self.cache.is_dirty()

    def flush(self):
        if self.fresh_bytez:
            warnings.warn("Buffer contains (probably) data which is not in the cache, it will be overwritten",
                          RuntimeWarning)

        data = self.cache.get_as_dict()
        bytez = self.block_definition.to_bytes(data)
        self.vulkan_buffer.map(bytez)
        self.cache.set_dirty(False)
        self.fresh_bytez = False

    def fetch(self):
        if self.cache.is_dirty():
            warnings.warn("Cache contains data which is not in the buffer, it will by overwritten", RuntimeWarning)

        with self.vulkan_buffer.mapped() as bytebuffer:
            bytez = bytebuffer[:]
        data = self.block_definition.from_bytes(bytez)
        self.cache.set_from_dict(data)
        self.cache.set_dirty(False)
        self.fresh_bytez = False


class BufferGPU(Buffer):

    def __init__(self, session, block_definition, block_usage):
        super(BufferGPU, self).__init__(session, block_definition, block_usage, Buffer.LOCATION_GPU)
        self.buffer_cpu = None
        self.copy_operation = CopyOperation(self.session.device, self.session.queue_index)

    @classmethod
    def from_shader(cls, session, shader, binding):
        block_definition = shader.get_block_definition(binding)
        block_usage = shader.get_block_usage(binding)
        return cls(session, block_definition, block_usage)

    def __getitem__(self, key):
        raise LavaUnsupportedError("Unsupported, the only way to read gpu buffers, is to copy them to a cpu buffer")

    def __setitem__(self, key, value):
        raise LavaUnsupportedError("Unsupported, the only way to write gpu buffers, is to copy from a cpu buffer")


class StagedBuffer(BufferInterface):

    def __init__(self, session, block_definition, block_usage, sync_mode=BufferInterface.SYNC_DEFAULT):
        super(StagedBuffer, self).__init__(session, block_definition, block_usage)
        self.buffer_cpu = BufferCPU(session, block_definition, block_usage, sync_mode)
        self.buffer_gpu = BufferGPU(session, block_definition, block_usage)
        self.sync_mode = sync_mode
        self.fresh_bytez = False

    def _destroy(self):
        self.buffer_cpu.destroy()
        self.buffer_gpu.destroy()

    @classmethod
    def from_shader(cls, session, shader, binding, sync_mode=BufferInterface.SYNC_DEFAULT):
        block_definition = shader.get_block_definition(binding)
        block_usage = shader.get_block_usage(binding)
        return cls(session, block_definition, block_usage, sync_mode)

    def get_vulkan_buffer(self):
        return self.buffer_gpu.get_vulkan_buffer()

    def before_stage(self, stage, binding, access_modifier):
        if access_modifier in [Access.READ_ONLY, Access.READ_WRITE]:
            if self.sync_mode is self.SYNC_LAZY and self.buffer_cpu.cache.is_dirty():
                self.flush()

    def after_stage(self, stage, binding, access_modifier):
        if access_modifier in [Access.WRITE_ONLY, Access.READ_WRITE]:
            self.fresh_bytez = True
            if self.sync_mode is self.SYNC_EAGER:
                self.fetch()

    def __getitem__(self, key):
        if self.fresh_bytez:
            self.fetch()
        return self.buffer_cpu[key]

    def __setitem__(self, key, value):
        self.buffer_cpu[key] = value

        if self.sync_mode is self.SYNC_EAGER:
            self.flush()

    def is_synced(self):
        return not self.fresh_bytez and self.buffer_cpu.is_synced()

    def flush(self):
        if self.buffer_cpu.cache.is_dirty():
            self.buffer_cpu.flush()
        self.buffer_cpu.copy_to(self.buffer_gpu)
        self.fresh_bytez = False

    def fetch(self):
        self.buffer_gpu.copy_to(self.buffer_cpu)
        self.buffer_cpu.fetch()
        self.fresh_bytez = False



# -*- coding: UTF-8 -*-

import contextlib

import vulkan as vk

from lava.api.constants.vk import BufferUsage, DescriptorType
from lava.api.util import Destroyable


class Memory(Destroyable):

    def __init__(self, device, memory_type_index, size):
        super(Memory, self).__init__()
        self.device = device
        self.memory_type_index = memory_type_index
        self.size = size
        self.memory_obj = None

        allocate_info = vk.VkMemoryAllocateInfo(allocationSize=size, memoryTypeIndex=memory_type_index)
        self.handle = vk.vkAllocateMemory(self.device.handle, allocate_info, None)

        # try:
        #     self.handle = vk.vkAllocateMemory(self.device.handle, allocate_info, None)
        # except vk.VkErrorOutOfDeviceMemory:
        #     raise RuntimeError("Device {} is out of memory".format(self.device.get_physical_device().get_name()))
        # except vk.VkErrorOutOfHostMemory:
        #     raise RuntimeError("Host is out of memory")
        # except vk.VkErrorTooManyObjects:
        #     raise RuntimeError("Too many allocations")

        # VK_ERROR_INVALID_EXTERNAL_HANDLE

    def _destroy(self):
        vk.vkFreeMemory(self.device.handle, self.handle, None)

    def get_size(self):
        return self.size

    def get_device(self):
        return self.device

    def get_memory_obj(self):
        return self.memory_obj

    @contextlib.contextmanager
    def mapped(self, size=-1, offset=0):
        if size == -1:
            size = self.size
        # https://cffi.readthedocs.io/en/latest/ref.html#ffi-buffer-ffi-from-buffer
        bytebuffer = vk.vkMapMemory(self.device.handle, self.handle, offset, size, flags=0)
        yield bytebuffer
        vk.vkUnmapMemory(self.device.handle, self.handle)


class MemoryObject(Destroyable):

    def __init__(self, device, size):
        super(MemoryObject, self).__init__()
        self.device = device
        self.size = size
        self.memory = None
        self.memory_offset = None

    def _destroy(self):
        raise NotImplementedError()

    def get_memory(self):
        return self.memory

    def get_size(self):
        return self.size

    def bind_memory(self, memory, memory_offset):
        if memory.get_memory_obj() is not None:
            raise RuntimeError("Memory is already bound")
        self.memory = memory
        self.memory_offset = memory_offset

    def descriptor_type(self):
        raise NotImplementedError()

    def descriptor_set_layout(self, binding):
        raise NotImplementedError()

    def write_descriptor_set(self, descriptor_set_handle, binding):
        raise NotImplementedError()


class Buffer(MemoryObject):

    def __init__(self, device, size, usage, queue_index):
        super(Buffer, self).__init__(device, size)
        self.usage = usage
        self.queue_index = queue_index

        usage_flags = BufferUsage.to_vulkan(usage)
        usage_flags |= vk.VK_BUFFER_USAGE_TRANSFER_SRC_BIT
        usage_flags |= vk.VK_BUFFER_USAGE_TRANSFER_DST_BIT

        create_info = vk.VkBufferCreateInfo(
            size=self.size, usage=usage_flags, sharingMode=vk.VK_SHARING_MODE_EXCLUSIVE,
            pQueueFamilyIndices=[queue_index]
        )

        self.handle = vk.vkCreateBuffer(self.device.handle, create_info, None)

    def _destroy(self):
        vk.vkDestroyBuffer(self.device.handle, self.handle, None)

    def get_memory_requirements(self):
        requirements = vk.vkGetBufferMemoryRequirements(self.device.handle, self.handle)

        supported_memory_indices = []
        for i in range(32):
            if requirements.memoryTypeBits & (1 << i) > 0:
                supported_memory_indices.append(i)

        return requirements.size, requirements.alignment, supported_memory_indices

    def bind_memory(self, memory, offset=0):
        super(Buffer, self).bind_memory(memory, offset)
        vk.vkBindBufferMemory(device=self.device.handle, buffer=self.handle, memory=self.memory.handle,
                              memoryOffset=offset)

    @contextlib.contextmanager
    def mapped(self):
        with self.memory.mapped(self.size, self.memory_offset) as bytebuffer:
            yield bytebuffer

    def map(self, bytez):
        with self.mapped() as bytebuffer:
            bytebuffer[:] = bytez

    def descriptor_type(self):
        if self.usage == BufferUsage.STORAGE_BUFFER:
            descriptor_type = DescriptorType.STORAGE_BUFFER
        elif self.usage == BufferUsage.UNIFORM_BUFFER:
            descriptor_type = DescriptorType.UNIFORM_BUFFER
        else:
            raise NotImplementedError("Buffer usage {} is not supported".format(self.usage))

        return descriptor_type

    def descriptor_set_layout(self, binding):
        descriptor_type = self.descriptor_type()
        return vk.VkDescriptorSetLayoutBinding(binding=binding,
                                               descriptorType=DescriptorType.to_vulkan(descriptor_type),
                                               descriptorCount=1, stageFlags=vk.VK_SHADER_STAGE_COMPUTE_BIT)

    def write_descriptor_set(self, descriptor_set_handle, binding):
        descriptor_type = self.descriptor_type()
        buffer_info = vk.VkDescriptorBufferInfo(self.handle, 0, self.size)
        return vk.VkWriteDescriptorSet(dstSet=descriptor_set_handle, dstBinding=binding,
                                       descriptorType=DescriptorType.to_vulkan(descriptor_type),
                                       pBufferInfo=[buffer_info])

# PushConstants ?
# SampledImage ?
# StorageImage ?



# -*- coding: UTF-8 -*-

import os
import platform
import subprocess

from lava import ENV_VAR_SDK

__all__ = ["compile_glsl"]

EXT_SPIR_V = ".spv"


def compile_glsl(path, verbose=True):
    path_output = path + EXT_SPIR_V

    if platform.system() not in ("Windows", "Linux"):
        raise NotImplementedError()

    if ENV_VAR_SDK not in os.environ:
        raise RuntimeError("Could not find environment variable {}".format(ENV_VAR_SDK))

    if platform.system() == "Linux":
        path_compiler = os.path.join(os.environ[ENV_VAR_SDK], "bin", "glslangValidator")
    else:
        path_compiler = os.path.join(os.environ[ENV_VAR_SDK], "Bin", "glslangValidator.exe")

    cmd = [path_compiler, "-V", path, "-o", path_output]

    if verbose:
        result = subprocess.call(cmd, stderr=subprocess.STDOUT)
    else:
        dev_null = open(os.devnull, "w")
        result = subprocess.call(cmd, stdout=dev_null, stderr=subprocess.STDOUT)
        dev_null.close()

    if result != 0:
        raise RuntimeError("Could not compile shader {}, try yourself with\n{}".format(path, " ".join(cmd)))

    return path_output



# -*- coding: UTF-8 -*-

from setuptools import setup

import lava as lv


setup(
    name="lava",
    version=lv.__version__,
    description="Highlevel Wrapper for Vulkan's Compute API",
    author="Jonas Schuepfer",
    author_email="jonasschuepfer@gmail.com",
    packages=["lava", "lava.api", "lava.api.constants"],
    include_package_data=True,
    install_requires=["vulkan", "numpy", "future"],
    url="https://github.com/osanj/lava",
    keywords=["Vulkan", "Parallel Computing", "Numpy"]
)



# -*- coding: UTF-8 -*-

import vulkan as vk

from lava.api.bytecode import ByteCode, ByteCodeError
from lava.api.bytes import Array, Matrix, Scalar, Struct, Vector
from lava.api.constants.spirv import Access, Decoration, ExecutionMode, ExecutionModel, Layout, Order, StorageClass
from lava.api.constants.vk import BufferUsage
from lava.api.util import Destroyable, LavaError, LavaUnsupportedError


class Shader(Destroyable):

    def __init__(self, device, path, entry_point=None):
        super(Shader, self).__init__()
        self.device = device
        with open(path, "rb") as f:
            self.bytez = f.read()
            self.handle = vk.vkCreateShaderModule(
                self.device.handle, vk.VkShaderModuleCreateInfo(codeSize=len(self.bytez), pCode=self.bytez), None)

        self.byte_code = ByteCode(self.bytez)

        # placeholder for inspection variables
        self.definitions_block = None
        self.block_data = None

        # check / set entry point
        self.entry_point, self.entry_point_index = self.check_entry_point(entry_point)
        self.local_size = self.check_local_size(self.entry_point_index)

    def _destroy(self):
        vk.vkDestroyShaderModule(self.device.handle, self.handle, None)

    def check_entry_point(self, entry_point):
        entry_points_detected = self.byte_code.find_entry_points(execution_model=ExecutionModel.GL_COMPUTE)
        index = None

        if len(entry_points_detected) == 0:
            raise RuntimeError("Could not find entry points for execution model {}".format(ExecutionModel.GL_COMPUTE))

        if entry_point is not None and entry_point not in entry_points_detected.values():
            raise RuntimeError("Could not find entry point {} in detected entry points {}".format(
                entry_point, ", ".join(entry_points_detected)))

        if entry_point is None:
            if len(entry_points_detected) > 1:
                raise RuntimeError("Multiple entry points found {}".format(", ".join(entry_points_detected.values())))
            entry_point = list(entry_points_detected.values())[0]

        for index_candidate, entry_point_candidate in entry_points_detected.items():
            if entry_point_candidate == entry_point:
                index = index_candidate

        return entry_point, index

    def check_local_size(self, entry_point_index):
        execution_mode, literals = self.byte_code.find_entry_point_details(entry_point_index)

        if execution_mode != ExecutionMode.LOCAL_SIZE:
            raise LavaUnsupportedError("Unsupported execution mode {}".format(execution_mode))

        return literals

    def get_entry_point(self):
        return self.entry_point

    def get_local_size(self):
        return self.local_size

    def inspect(self):
        self.block_data = self.byte_code.find_blocks()
        self.definitions_block = {}

        defs_scalar = {index: Scalar.of(dtype) for index, dtype in self.byte_code.types_scalar.items()}
        defs_vector = {index: Vector(n, dtype) for index, (dtype, n) in self.byte_code.types_vector.items()}
        defs_array = {}
        defs_struct = {}

        for binding in self.get_bindings():
            index, _ = self.get_block_index(binding)
            self.deduce_definition(index, defs_scalar, defs_vector, defs_array, defs_struct, index)
            self.definitions_block[index] = defs_struct[index]
            self.deduce_layout(binding)

    def deduce_definition(self, index, definitions_scalar, definitions_vector, definitions_array, definitions_struct,
                          last_struct=None):
        default_layout = Layout.STD140

        if index in self.byte_code.types_array:
            type_index, dims = self.byte_code.types_array[index]

            # build missing definition
            if type_index in self.byte_code.types_struct and type_index not in definitions_struct:
                self.deduce_definition(type_index, definitions_scalar, definitions_vector, definitions_array,
                                       definitions_struct, last_struct)

            definition = None

            # matrix types are shared, but still affected by the layout, create a instance for every occurrence
            if type_index in self.byte_code.types_matrix:
                definition = self.build_matrix_definition(type_index, default_layout, last_struct)

            definition = definition or definitions_scalar.get(type_index)
            definition = definition or definitions_vector.get(type_index)
            definition = definition or definitions_struct.get(type_index)

            definitions_array[index] = Array(definition.copy(), dims, default_layout)

        elif index in self.byte_code.types_struct:
            member_indices = self.byte_code.types_struct[index]

            # build missing definitions
            for member_index in member_indices:
                is_struct = member_index in self.byte_code.types_struct
                is_array = member_index in self.byte_code.types_array

                defs = {
                    "definitions_scalar": definitions_scalar,
                    "definitions_vector": definitions_vector,
                    "definitions_array": definitions_array,
                    "definitions_struct": definitions_struct
                }

                if is_struct and member_index not in definitions_struct:
                    self.deduce_definition(member_index, last_struct=member_index, **defs)

                if is_array and member_index not in definitions_array:
                    self.deduce_definition(member_index, last_struct=index, **defs)

            definitions = []
            for member_index in member_indices:
                definition = None

                # matrix types are shared, but still affected by the layout, create a instance for every occurrence
                if member_index in self.byte_code.types_matrix:
                    definition = self.build_matrix_definition(member_index, default_layout, last_struct)

                definition = definition or definitions_scalar.get(member_index)
                definition = definition or definitions_vector.get(member_index)
                definition = definition or definitions_array.get(member_index)
                definition = definition or definitions_struct.get(member_index)
                definitions.append(definition.copy())

            struct_name, member_names = self.byte_code.find_names(index)
            definitions_struct[index] = Struct(definitions, default_layout, member_names=member_names,
                                               type_name=struct_name)

        else:
            raise ByteCodeError.unexpected()

    def build_matrix_definition(self, matrix_index, layout, last_struct_index):
        # we assume that matrices always have the order decoration in bytecode (if they are part of a interface block)
        if last_struct_index is None:
            raise ByteCodeError.unexpected()

        member_indices = self.byte_code.find_member_ids(last_struct_index)
        member_orders = self.byte_code.find_orders(last_struct_index)

        # matrix is direct member of the last struct
        if matrix_index in member_indices:
            member = member_indices.index(matrix_index)

        # if the matrix is a member of an array which is a struct member, the array is also decorated with the
        # matrix order
        else:
            member = None

            for index in member_indices:
                if index in self.byte_code.types_array:
                    type_index, _ = self.byte_code.types_array[index]
                    if type_index == matrix_index:
                        member = member_indices.index(index)
                        break

            if member is None:
                raise ByteCodeError.unexpected()

        order_decoration = member_orders[member]
        order = {Decoration.ROW_MAJOR: Order.ROW_MAJOR, Decoration.COL_MAJOR: Order.COLUMN_MAJOR}[order_decoration]

        dtype, rows, cols = self.byte_code.types_matrix[matrix_index]
        return Matrix(cols, rows, dtype, layout, order)

    def deduce_layout(self, binding):
        index, usage = self.get_block_index(binding)
        definition = self.get_block_definition(binding)

        definition.layout = Layout.STD140
        match_std140 = self.check_layout(index)

        definition.layout = Layout.STD430
        match_std430 = self.check_layout(index)

        # deduce layout
        if match_std140 and not match_std430:
            definition.layout = Layout.STD140

        elif not match_std140 and match_std430:
            definition.layout = Layout.STD430

        elif match_std140 and match_std430:
            # std430 is not allowed for uniform buffer objects
            if usage == BufferUsage.UNIFORM_BUFFER:
                definition.layout = Layout.STD140
            else:
                definition.layout = Layout.STDXXX

        else:
            possible_reasons = [
                "a memory layout other than std140 or std430 was used",
                "an offset was defined manually, e.g. for a struct member: layout(offset=128) int member;",
                "a matrix memory order was defined manually, e.g. for a struct member: layout(row_major) "
                "StructXYZ structWithMatrices;"
            ]
            raise RuntimeError("Found unexpected memory offsets, this might occur because of\n" +
                               "".join(["* {}\n".format(reason) for reason in possible_reasons]))

    def check_layout(self, index):
        definition = self.definitions_block[index]

        member_indices = self.byte_code.find_member_ids(index)
        offsets_bytecode = self.byte_code.find_offsets(index)
        offsets_bytecode = [offsets_bytecode.get(i) for i in range(len(definition.definitions))]

        if None in offsets_bytecode:
            raise ByteCodeError.unexpected()

        if offsets_bytecode != definition.offsets():
            return False

        for i, (member_index, d) in enumerate(zip(member_indices, definition.definitions)):
            if isinstance(d, Array):
                if self.byte_code.find_strides(member_index) != d.strides():
                    return False

            if isinstance(d, Matrix):
                if self.byte_code.find_matrix_stride(index, i) != d.stride():
                    return False

        return True

    def get_bindings(self):
        bindings = []

        for index in self.block_data:
            bindings.append(self.block_data[index][2])

        for binding in bindings:
            count = bindings.count(binding)
            if count > 1:
                raise LavaError("Binding {} is used {} times".format(binding, count))

        return list(sorted(bindings))

    def get_block_index(self, binding):
        for index in self.block_data:
            block_type, storage_class, binding_id = self.block_data[index]

            if binding_id == binding:
                usage = None

                if block_type == Decoration.BLOCK and storage_class == StorageClass.UNIFORM:
                    usage = BufferUsage.UNIFORM_BUFFER
                elif block_type == Decoration.BUFFER_BLOCK and storage_class == StorageClass.UNIFORM:
                    usage = BufferUsage.STORAGE_BUFFER

                return index, usage

        raise ValueError("Binding {} not found".format(binding))

    def get_block(self, binding):
        index, usage = self.get_block_index(binding)
        return self.definitions_block[index], usage

    def get_block_definition(self, binding):
        return self.get_block(binding)[0]

    def get_block_usage(self, binding):
        return self.get_block(binding)[1]

    def get_block_access(self, binding):
        index, usage = self.get_block_index(binding)
        block_definition, _ = self.get_block(binding)

        if usage == BufferUsage.UNIFORM_BUFFER:
            return Access.READ_ONLY

        decorations = self.byte_code.find_accesses(index)

        if len(decorations) not in (0, len(block_definition.definitions)):  # 0 for no access decorations at all
            raise ByteCodeError.unexpected()

        accesses = set()

        for member, d in decorations.items():
            if Decoration.NON_WRITABLE in d and Decoration.NON_READABLE not in d:
                accesses.add(Access.READ_ONLY)
            elif Decoration.NON_WRITABLE not in d and Decoration.NON_READABLE in d:
                accesses.add(Access.WRITE_ONLY)
            elif Decoration.NON_WRITABLE in d and Decoration.NON_READABLE in d:
                accesses.add(Access.NEITHER)
            else:
                accesses.add(Access.READ_WRITE)

        if len(accesses) == 0:
            accesses.add(Access.READ_WRITE)
        elif len(accesses) > 1:
            raise ByteCodeError.unexpected()

        return accesses.pop()



